from __future__ import annotations

import asyncio
import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from infrahub_sdk.protocols import CoreArtifactValidator, CoreGeneratorDefinition, CoreProposedChange
from prefect import flow, task
from prefect.cache_policies import NONE
from prefect.client.schemas.objects import (
    State,  # noqa: TC002
)
from prefect.logging import get_run_logger
from prefect.states import Completed, Failed

from infrahub import config
from infrahub.artifacts.models import CheckArtifactCreate
from infrahub.context import InfrahubContext  # noqa: TC001  needed for prefect flow
from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.branch.tasks import merge_branch
from infrahub.core.constants import InfrahubKind, RepositoryInternalStatus, ValidatorConclusion
from infrahub.core.diff.coordinator import DiffCoordinator
from infrahub.core.diff.model.diff import DiffElementType, SchemaConflict
from infrahub.core.diff.model.path import NodeDiffFieldSummary
from infrahub.core.integrity.object_conflict.conflict_recorder import ObjectConflictValidatorRecorder
from infrahub.core.protocols import CoreDataCheck, CoreValidator
from infrahub.core.protocols import CoreProposedChange as InternalCoreProposedChange
from infrahub.core.validators.checks_runner import run_checks_and_update_validator
from infrahub.core.validators.determiner import ConstraintValidatorDeterminer
from infrahub.core.validators.models.validate_migration import SchemaValidateMigrationData
from infrahub.core.validators.tasks import schema_validate_migrations
from infrahub.dependencies.registry import get_component_registry
from infrahub.exceptions import MergeFailedError
from infrahub.generators.models import ProposedChangeGeneratorDefinition
from infrahub.git.models import TriggerRepositoryInternalChecks, TriggerRepositoryUserChecks
from infrahub.git.repository import get_initialized_repo
from infrahub.log import get_logger
from infrahub.message_bus import InfrahubMessage, messages
from infrahub.message_bus.operations.requests.proposed_change import DefinitionSelect
from infrahub.proposed_change.constants import ProposedChangeState
from infrahub.proposed_change.models import (
    RequestArtifactDefinitionCheck,
    RequestProposedChangeDataIntegrity,
    RequestProposedChangeRepositoryChecks,
    RequestProposedChangeRunGenerators,
    RequestProposedChangeSchemaIntegrity,
    RequestProposedChangeUserTests,
)
from infrahub.pytest_plugin import InfrahubBackendPlugin
from infrahub.services import InfrahubServices  # noqa: TC001  needed for prefect flow
from infrahub.validators.tasks import start_validator
from infrahub.workflows.catalogue import (
    GIT_REPOSITORIES_CHECK_ARTIFACT_CREATE,
    GIT_REPOSITORY_INTERNAL_CHECKS_TRIGGER,
    GIT_REPOSITORY_USER_CHECKS_TRIGGER,
    REQUEST_PROPOSED_CHANGE_REPOSITORY_CHECKS,
)
from infrahub.workflows.utils import add_tags

from .branch_diff import get_diff_summary_cache, get_modified_kinds

if TYPE_CHECKING:
    from infrahub_sdk.diff import NodeDiff
    from infrahub_sdk.node import InfrahubNode

    from infrahub.core.models import SchemaUpdateConstraintInfo
    from infrahub.core.schema.schema_branch import SchemaBranch
    from infrahub.message_bus.types import ProposedChangeRepository


async def _proposed_change_transition_state(
    state: ProposedChangeState,
    service: InfrahubServices,
    proposed_change: InternalCoreProposedChange | None = None,
    proposed_change_id: str | None = None,
) -> None:
    async with service.database.start_session() as db:
        if proposed_change is None and proposed_change_id:
            proposed_change = await registry.manager.get_one(
                db=db, id=proposed_change_id, kind=InternalCoreProposedChange, raise_on_error=True
            )
        if proposed_change:
            proposed_change.state.value = state.value  # type: ignore[misc]
            await proposed_change.save(db=db)


# async def proposed_change_transition_merged(flow: Flow, flow_run: FlowRun, state: State) -> None:
#     await _proposed_change_transition_state(
#         proposed_change_id=flow_run.parameters["proposed_change_id"], state=ProposedChangeState.MERGED
#     )


# async def proposed_change_transition_open(flow: Flow, flow_run: FlowRun, state: State) -> None:
#     await _proposed_change_transition_state(
#         proposed_change_id=flow_run.parameters["proposed_change_id"], state=ProposedChangeState.OPEN
#     )


@flow(
    name="proposed-change-merge",
    flow_run_name="Merge propose change: {proposed_change_name} ",
    description="Merge a given proposed change.",
    # TODO need to investigate why these function are not working as expected
    # on_completion=[proposed_change_transition_merged],  # type: ignore
    # on_failure=[proposed_change_transition_open],  # type: ignore
    # on_crashed=[proposed_change_transition_open],  # type: ignore
    # on_cancellation=[proposed_change_transition_open],  # type: ignore
)
async def merge_proposed_change(
    proposed_change_id: str,
    proposed_change_name: str,  # noqa: ARG001
    context: InfrahubContext,
    service: InfrahubServices,
) -> State:
    log = get_run_logger()

    await add_tags(nodes=[proposed_change_id])

    async with service.database.start_session() as db:
        proposed_change = await registry.manager.get_one(
            db=db, id=proposed_change_id, kind=InternalCoreProposedChange, raise_on_error=True
        )

        log.info("Validating if all conditions are met to merge the proposed change")

        source_branch = await Branch.get_by_name(db=db, name=proposed_change.source_branch.value)
        validations = await proposed_change.validations.get_peers(db=db, peer_type=CoreValidator)
        for validation in validations.values():
            validator_kind = validation.get_kind()
            if (
                validator_kind != InfrahubKind.DATAVALIDATOR
                and validation.conclusion.value.value != ValidatorConclusion.SUCCESS.value
            ):
                # Ignoring Data integrity checks as they are handled again later
                await _proposed_change_transition_state(
                    proposed_change=proposed_change, state=ProposedChangeState.OPEN, service=service
                )
                return Failed(message="Unable to merge proposed change containing failing checks")
            if validator_kind == InfrahubKind.DATAVALIDATOR:
                data_checks = await validation.checks.get_peers(db=db, peer_type=CoreDataCheck)
                for check in data_checks.values():
                    if check.conflicts.value and not check.keep_branch.value:
                        await _proposed_change_transition_state(
                            proposed_change=proposed_change, state=ProposedChangeState.OPEN, service=service
                        )
                        return Failed(
                            message="Data conflicts found on branch and missing decisions about what branch to keep"
                        )

        log.info("Proposed change is eligible to be merged")
        try:
            await merge_branch(
                branch=source_branch.name, context=context, service=service, proposed_change_id=proposed_change_id
            )
        except MergeFailedError as exc:
            await _proposed_change_transition_state(
                proposed_change=proposed_change, state=ProposedChangeState.OPEN, service=service
            )
            return Failed(message=f"Merge failure when trying to merge {exc.message}")

        log.info(f"Branch {source_branch.name} has been merged successfully")

        await _proposed_change_transition_state(
            proposed_change=proposed_change, state=ProposedChangeState.MERGED, service=service
        )
        return Completed(message="proposed change merged successfully")


@flow(
    name="proposed-changes-cancel-branch",
    flow_run_name="Cancel all proposed change associated with branch {branch_name}",
    description="Cancel all Proposed change associated with a branch.",
)
async def cancel_proposed_changes_branch(branch_name: str, service: InfrahubServices) -> None:
    await add_tags(branches=[branch_name])

    proposed_changed_opened = await service.client.filters(
        kind=CoreProposedChange,
        include=["id", "source_branch"],
        state__value=ProposedChangeState.OPEN.value,
        source_branch__value=branch_name,
    )
    proposed_changed_closed = await service.client.filters(
        kind=CoreProposedChange,
        include=["id", "source_branch"],
        state__value=ProposedChangeState.CLOSED.value,
        source_branch__value=branch_name,
    )

    for proposed_change in proposed_changed_opened + proposed_changed_closed:
        await cancel_proposed_change(proposed_change=proposed_change, service=service)


@task(name="Cancel a propose change", description="Cancel a propose change", cache_policy=NONE)  # type: ignore[arg-type]
async def cancel_proposed_change(proposed_change: CoreProposedChange, service: InfrahubServices) -> None:
    await add_tags(nodes=[proposed_change.id])
    log = get_run_logger()

    log.info("Canceling proposed change as the source branch was deleted")
    proposed_change = await service.client.get(kind=CoreProposedChange, id=proposed_change.id)
    proposed_change.state.value = ProposedChangeState.CANCELED.value
    await proposed_change.save()


@flow(
    name="proposed-changed-data-integrity",
    flow_run_name="Triggers data integrity check",
)
async def run_proposed_change_data_integrity_check(
    model: RequestProposedChangeDataIntegrity, service: InfrahubServices
) -> None:
    """Triggers a data integrity validation check on the provided proposed change to start."""
    await add_tags(branches=[model.source_branch], nodes=[model.proposed_change])

    async with service.database.start_transaction() as dbt:
        destination_branch = await registry.get_branch(db=dbt, branch=model.destination_branch)
        source_branch = await registry.get_branch(db=dbt, branch=model.source_branch)
        component_registry = get_component_registry()

        diff_coordinator = await component_registry.get_component(DiffCoordinator, db=dbt, branch=source_branch)
        await diff_coordinator.update_branch_diff(base_branch=destination_branch, diff_branch=source_branch)


@flow(
    name="proposed-changed-run-generator",
    flow_run_name="Run generators",
)
async def run_generators(
    model: RequestProposedChangeRunGenerators, context: InfrahubContext, service: InfrahubServices
) -> None:
    await add_tags(branches=[model.source_branch], nodes=[model.proposed_change], db_change=True)

    generators = await service.client.filters(
        kind=CoreGeneratorDefinition,
        prefetch_relationships=True,
        populate_store=True,
        branch=model.source_branch,
    )
    generator_definitions = [
        ProposedChangeGeneratorDefinition(
            definition_id=generator.id,
            definition_name=generator.name.value,
            class_name=generator.class_name.value,
            file_path=generator.file_path.value,
            query_name=generator.query.peer.name.value,
            query_models=generator.query.peer.models.value,
            repository_id=generator.repository.peer.id,
            parameters=generator.parameters.value,
            group_id=generator.targets.peer.id,
            convert_query_response=generator.convert_query_response.value,
        )
        for generator in generators
    ]

    diff_summary = await get_diff_summary_cache(pipeline_id=model.branch_diff.pipeline_id, cache=service.cache)
    modified_kinds = get_modified_kinds(diff_summary=diff_summary, branch=model.source_branch)

    for generator_definition in generator_definitions:
        # Request generator definitions if the source branch that is managed in combination
        # to the Git repository containing modifications which could indicate changes to the transforms
        # in code
        # Alternatively if the queries used touches models that have been modified in the path
        # impacted artifact definitions will be included for consideration

        select = DefinitionSelect.NONE
        select = select.add_flag(
            current=select,
            flag=DefinitionSelect.FILE_CHANGES,
            condition=model.source_branch_sync_with_git and model.branch_diff.has_file_modifications,
        )

        for changed_model in modified_kinds:
            select = select.add_flag(
                current=select,
                flag=DefinitionSelect.MODIFIED_KINDS,
                condition=changed_model in generator_definition.query_models,
            )

        if select:
            msg = messages.RequestGeneratorDefinitionCheck(
                context=context,
                generator_definition=generator_definition,
                branch_diff=model.branch_diff,
                proposed_change=model.proposed_change,
                source_branch=model.source_branch,
                source_branch_sync_with_git=model.source_branch_sync_with_git,
                destination_branch=model.destination_branch,
            )
            msg.assign_meta(parent=model)
            await service.message_bus.send(message=msg)

    next_messages: list[InfrahubMessage] = []
    if model.refresh_artifacts:
        next_messages.append(
            messages.RequestProposedChangeRefreshArtifacts(
                context=context,
                proposed_change=model.proposed_change,
                source_branch=model.source_branch,
                source_branch_sync_with_git=model.source_branch_sync_with_git,
                destination_branch=model.destination_branch,
                branch_diff=model.branch_diff,
            )
        )

    if model.do_repository_checks:
        model_proposed_change_repo_checks = RequestProposedChangeRepositoryChecks(
            proposed_change=model.proposed_change,
            source_branch=model.source_branch,
            source_branch_sync_with_git=model.source_branch_sync_with_git,
            destination_branch=model.destination_branch,
            branch_diff=model.branch_diff,
        )
        await service.workflow.submit_workflow(
            workflow=REQUEST_PROPOSED_CHANGE_REPOSITORY_CHECKS,
            context=context,
            parameters={"model": model_proposed_change_repo_checks},
        )

    for next_msg in next_messages:
        next_msg.assign_meta(parent=model)
        await service.message_bus.send(message=next_msg)


@flow(
    name="proposed-changed-schema-integrity",
    flow_run_name="Process schema integrity",
)
async def run_proposed_change_schema_integrity_check(
    model: RequestProposedChangeSchemaIntegrity, service: InfrahubServices
) -> None:
    # For now, we retrieve the latest schema for each branch from the registry
    # In the future it would be good to generate the object SchemaUpdateValidationResult from message.branch_diff
    await add_tags(branches=[model.source_branch], nodes=[model.proposed_change])

    source_schema = registry.schema.get_schema_branch(name=model.source_branch).duplicate()
    dest_schema = registry.schema.get_schema_branch(name=model.destination_branch).duplicate()

    candidate_schema = dest_schema.duplicate()
    candidate_schema.update(schema=source_schema)
    schema_diff = dest_schema.diff(other=candidate_schema)
    validation_result = dest_schema.validate_update(other=candidate_schema, diff=schema_diff)

    diff_summary = await get_diff_summary_cache(pipeline_id=model.branch_diff.pipeline_id, cache=service.cache)
    constraints_from_data_diff = await _get_proposed_change_schema_integrity_constraints(
        schema=candidate_schema, diff_summary=diff_summary
    )
    constraints_from_schema_diff = validation_result.constraints
    constraints = set(constraints_from_data_diff + constraints_from_schema_diff)

    if not constraints:
        return

    # ----------------------------------------------------------
    # Validate if the new schema is valid with the content of the database
    # ----------------------------------------------------------
    source_branch = registry.get_branch_from_registry(branch=model.source_branch)
    responses = await schema_validate_migrations(
        message=SchemaValidateMigrationData(
            branch=source_branch, schema_branch=candidate_schema, constraints=list(constraints)
        ),
        service=service,
    )

    # TODO we need to report a failure if an error happened during the execution of a validator
    conflicts: list[SchemaConflict] = []
    for response in responses:
        for violation in response.violations:
            conflicts.append(
                SchemaConflict(
                    name=response.schema_path.get_path(),
                    type=response.constraint_name,
                    kind=violation.node_kind,
                    id=violation.node_id,
                    path=response.schema_path.get_path(),
                    value=violation.message,
                    branch="placeholder",
                )
            )

    if not conflicts:
        return

    async with service.database.start_transaction() as db:
        object_conflict_validator_recorder = ObjectConflictValidatorRecorder(
            db=db,
            validator_kind=InfrahubKind.SCHEMAVALIDATOR,
            validator_label="Schema Integrity",
            check_schema_kind=InfrahubKind.SCHEMACHECK,
        )
        await object_conflict_validator_recorder.record_conflicts(
            proposed_change_id=model.proposed_change, conflicts=conflicts
        )


async def _get_proposed_change_schema_integrity_constraints(
    schema: SchemaBranch, diff_summary: list[NodeDiff]
) -> list[SchemaUpdateConstraintInfo]:
    node_diff_field_summary_map: dict[str, NodeDiffFieldSummary] = {}

    for node_diff in diff_summary:
        node_kind = node_diff["kind"]
        if node_kind not in node_diff_field_summary_map:
            node_diff_field_summary_map[node_kind] = NodeDiffFieldSummary(kind=node_kind)
        field_summary = node_diff_field_summary_map[node_kind]
        for element in node_diff["elements"]:
            element_name = element["name"]
            element_type = element["element_type"]
            if element_type.lower() in (
                DiffElementType.RELATIONSHIP_MANY.value.lower(),
                DiffElementType.RELATIONSHIP_ONE.value.lower(),
            ):
                field_summary.relationship_names.add(element_name)
            elif element_type.lower() in (DiffElementType.ATTRIBUTE.value.lower(),):
                field_summary.attribute_names.add(element_name)

    determiner = ConstraintValidatorDeterminer(schema_branch=schema)
    return await determiner.get_constraints(node_diffs=list(node_diff_field_summary_map.values()))


@flow(
    name="proposed-changed-repository-checks",
    flow_run_name="Process user defined checks",
)
async def repository_checks(
    model: RequestProposedChangeRepositoryChecks, service: InfrahubServices, context: InfrahubContext
) -> None:
    await add_tags(branches=[model.source_branch], nodes=[model.proposed_change])

    for repository in model.branch_diff.repositories:
        if (
            model.source_branch_sync_with_git
            and not repository.read_only
            and repository.internal_status == RepositoryInternalStatus.ACTIVE.value
        ):
            trigger_internal_checks_model = TriggerRepositoryInternalChecks(
                proposed_change=model.proposed_change,
                repository=repository.repository_id,
                source_branch=model.source_branch,
                target_branch=model.destination_branch,
            )
            await service.workflow.submit_workflow(
                workflow=GIT_REPOSITORY_INTERNAL_CHECKS_TRIGGER,
                context=context,
                parameters={"model": trigger_internal_checks_model},
            )

        trigger_user_checks_model = TriggerRepositoryUserChecks(
            proposed_change=model.proposed_change,
            repository_id=repository.repository_id,
            repository_name=repository.repository_name,
            source_branch=model.source_branch,
            source_branch_sync_with_git=model.source_branch_sync_with_git,
            target_branch=model.destination_branch,
            branch_diff=model.branch_diff,
        )
        await service.workflow.submit_workflow(
            workflow=GIT_REPOSITORY_USER_CHECKS_TRIGGER,
            context=context,
            parameters={"model": trigger_user_checks_model},
        )


@flow(
    name="proposed-changed-user-tests",
    flow_run_name="Run unit tests in repositories",
)
async def run_proposed_change_user_tests(model: RequestProposedChangeUserTests, service: InfrahubServices) -> None:
    log = get_run_logger()
    await add_tags(branches=[model.source_branch], nodes=[model.proposed_change])
    proposed_change = await service.client.get(kind=InfrahubKind.PROPOSEDCHANGE, id=model.proposed_change)

    def _execute(
        directory: Path, repository: ProposedChangeRepository, proposed_change: InfrahubNode
    ) -> int | pytest.ExitCode:
        config_file = str(directory / ".infrahub.yml")
        test_directory = directory / "tests"
        log = get_logger()

        if not test_directory.is_dir():
            log.debug(
                event="repository_tests_ignored",
                proposed_change=proposed_change,
                repository=repository.repository_name,
                message="tests directory not found",
            )
            return 1

        # Redirect stdout/stderr to avoid showing pytest lines in the git agent
        old_out = sys.stdout
        old_err = sys.stderr

        with Path(os.devnull).open(mode="w", encoding="utf-8") as devnull:
            sys.stdout = devnull
            sys.stderr = devnull

            exit_code = pytest.main(
                [
                    str(test_directory),
                    f"--infrahub-repo-config={config_file}",
                    f"--infrahub-address={config.SETTINGS.main.internal_address}",
                    "-qqqq",
                    "-s",
                ],
                plugins=[InfrahubBackendPlugin(service.client.config, repository.repository_id, proposed_change.id)],
            )

        # Restore stdout/stderr back to their orignal states
        sys.stdout = old_out
        sys.stderr = old_err

        return exit_code

    for repository in model.branch_diff.repositories:
        if model.source_branch_sync_with_git:
            repo = await get_initialized_repo(
                repository_id=repository.repository_id,
                name=repository.repository_name,
                service=service,
                repository_kind=repository.kind,
            )
            commit = repo.get_commit_value(proposed_change.source_branch.value)
            worktree_directory = Path(repo.get_commit_worktree(commit=commit).directory)

            return_code = await asyncio.to_thread(_execute, worktree_directory, repository, proposed_change)
            log.info(msg=f"repository_tests_completed return_code={return_code}")


@flow(
    name="artifacts-generation-validation",
    flow_run_name="Validating generation of artifacts for {model.artifact_definition.definition_name}",
)
async def validate_artifacts_generation(model: RequestArtifactDefinitionCheck, service: InfrahubServices) -> None:
    await add_tags(branches=[model.source_branch], nodes=[model.proposed_change], db_change=True)

    log = get_run_logger()
    artifact_definition = await service.client.get(
        kind=InfrahubKind.ARTIFACTDEFINITION,
        id=model.artifact_definition.definition_id,
        branch=model.source_branch,
    )
    proposed_change = await service.client.get(kind=InfrahubKind.PROPOSEDCHANGE, id=model.proposed_change)

    validator_name = f"Artifact Validator: {model.artifact_definition.definition_name}"

    await proposed_change.validations.fetch()

    previous_validator: CoreArtifactValidator | None = None
    for relationship in proposed_change.validations.peers:
        existing_validator = relationship.peer
        if (
            existing_validator.typename == InfrahubKind.ARTIFACTVALIDATOR
            and existing_validator.definition.id == model.artifact_definition.definition_id
        ):
            previous_validator = existing_validator

    validator = await start_validator(
        service=service,
        validator=previous_validator,
        validator_type=CoreArtifactValidator,
        proposed_change=model.proposed_change,
        data={
            "label": validator_name,
            "definition": model.artifact_definition.definition_id,
        },
        context=model.context,
    )

    await artifact_definition.targets.fetch()
    group = artifact_definition.targets.peer
    await group.members.fetch()

    existing_artifacts = await service.client.filters(
        kind=InfrahubKind.ARTIFACT,
        definition__ids=[model.artifact_definition.definition_id],
        include=["object"],
        branch=model.source_branch,
    )
    artifacts_by_member = {}
    for artifact in existing_artifacts:
        artifacts_by_member[artifact.object.peer.id] = artifact.id

    repository = model.branch_diff.get_repository(repository_id=model.artifact_definition.repository_id)
    impacted_artifacts = model.branch_diff.get_subscribers_ids(kind=InfrahubKind.ARTIFACT)

    checks = []

    for relationship in group.members.peers:
        member = relationship.peer
        artifact_id = artifacts_by_member.get(member.id)
        if _should_render_artifact(
            artifact_id=artifact_id,
            managed_branch=model.source_branch_sync_with_git,
            impacted_artifacts=impacted_artifacts,
        ):
            log.info(f"Trigger Artifact processing for {member.display_label}")

            check_model = CheckArtifactCreate(
                context=model.context,
                artifact_name=model.artifact_definition.artifact_name,
                artifact_id=artifact_id,
                artifact_definition=model.artifact_definition.definition_id,
                commit=repository.source_commit,
                content_type=model.artifact_definition.content_type,
                transform_type=model.artifact_definition.transform_kind,
                transform_location=model.artifact_definition.transform_location,
                repository_id=repository.repository_id,
                repository_name=repository.repository_name,
                repository_kind=repository.kind,
                branch_name=model.source_branch,
                query=model.artifact_definition.query_name,
                variables=member.extract(params=artifact_definition.parameters.value),
                target_id=member.id,
                target_kind=member.get_kind(),
                target_name=member.display_label,
                timeout=model.artifact_definition.timeout,
                validator_id=validator.id,
            )

            checks.append(
                service.workflow.execute_workflow(
                    workflow=GIT_REPOSITORIES_CHECK_ARTIFACT_CREATE,
                    parameters={"model": check_model},
                    expected_return=ValidatorConclusion,
                )
            )

    await run_checks_and_update_validator(
        checks=checks,
        validator=validator,
        proposed_change_id=model.proposed_change,
        context=model.context,
        service=service,
    )


def _should_render_artifact(artifact_id: str | None, managed_branch: bool, impacted_artifacts: list[str]) -> bool:  # noqa: ARG001
    """Returns a boolean to indicate if an artifact should be generated or not.
    Will return true if:
        * The artifact_id wasn't set which could be that it's a new object that doesn't have a previous artifact
        * The source brance is not data only which would indicate that it could contain updates in git to the transform
        * The artifact_id exists in the impacted_artifacts list
    Will return false if:
        * The source branch is a data only branch and the artifact_id exists and is not in the impacted list
    """

    # if not artifact_id or managed_branch:
    #    return True
    # return artifact_id in impacted_artifacts
    # Temporary workaround tracked in https://github.com/opsmill/infrahub/issues/4991
    return True



import logging
from pathlib import Path

import typer
from rich.console import Console

from ..async_typer import AsyncTyper
from ..ctl.client import initialize_client
from ..ctl.utils import catch_exception, init_logging
from ..spec.menu import MenuFile
from .parameters import CONFIG_PARAM
from .utils import load_yamlfile_from_disk_and_exit

app = AsyncTyper()
console = Console()


@app.callback()
def callback() -> None:
    """
    Manage the menu in a remote Infrahub instance.
    """


@app.command()
@catch_exception(console=console)
async def load(
    menus: list[Path],
    debug: bool = False,
    branch: str = typer.Option(None, help="Branch on which to load the menu."),
    _: str = CONFIG_PARAM,
) -> None:
    """Load one or multiple menu files into Infrahub."""

    init_logging(debug=debug)

    logging.getLogger("infrahub_sdk").setLevel(logging.INFO)

    files = load_yamlfile_from_disk_and_exit(paths=menus, file_type=MenuFile, console=console)
    client = initialize_client()

    for file in files:
        file.validate_content()
        schema = await client.schema.get(kind=file.spec.kind, branch=branch)

        for idx, item in enumerate(file.spec.data):
            await file.spec.create_node(
                client=client,
                schema=schema,
                data=item,
                branch=branch,
                default_schema_kind=file.spec.kind,
                context={"list_index": idx},
            )



from rich.console import Console

from ..schema.repository import InfrahubRepositoryConfig
from ..template.exceptions import (
    JinjaTemplateError,
    JinjaTemplateNotFoundError,
    JinjaTemplateSyntaxError,
    JinjaTemplateUndefinedError,
)


def list_jinja2_transforms(config: InfrahubRepositoryConfig) -> None:
    console = Console()
    console.print(f"Jinja2 transforms defined in repository: {len(config.jinja2_transforms)}")

    for transform in config.jinja2_transforms:
        console.print(f"{transform.name} ({transform.template_path})")


def print_template_errors(error: JinjaTemplateError, console: Console) -> None:
    if isinstance(error, JinjaTemplateNotFoundError):
        console.print("[red]An error occurred while rendering the jinja template")
        console.print("")
        if error.base_template:
            console.print(f"Base template: [yellow]{error.base_template}")
        console.print(f"Missing template: [yellow]{error.filename}")
        return

    if isinstance(error, JinjaTemplateUndefinedError):
        console.print("[red]An error occurred while rendering the jinja template")
        for current_error in error.errors:
            console.print(f"[yellow]{current_error.frame.filename} on line {current_error.frame.lineno}\n")
            console.print(current_error.syntax)
        console.print("")
        console.print(error.message)
        return

    if isinstance(error, JinjaTemplateSyntaxError):
        console.print("[red]A syntax error was encountered within the template")
        console.print("")
        if error.filename:
            console.print(f"Filename: [yellow]{error.filename}")
        console.print(f"Line number: [yellow]{error.lineno}")
        console.print()
        console.print(error.message)
        return

    console.print("[red]An error occurred while rendering the jinja template")
    console.print("")
    console.print(f"[yellow]{error.message}")



from __future__ import annotations

import inspect
from collections import defaultdict
from dataclasses import dataclass
from typing import TYPE_CHECKING, Generator

from infrahub_sdk.uuidt import UUIDT

from infrahub.core.changelog.models import (
    ChangelogRelationshipMapper,
    RelationshipCardinalityManyChangelog,
    RelationshipCardinalityOneChangelog,
)
from infrahub.core.constants import RelationshipDirection, RelationshipStatus
from infrahub.core.constants.database import DatabaseEdgeType
from infrahub.core.query import Query, QueryType
from infrahub.core.query.subquery import build_subquery_filter, build_subquery_order
from infrahub.core.timestamp import Timestamp
from infrahub.core.utils import extract_field_filters
from infrahub.log import get_logger

if TYPE_CHECKING:
    from uuid import UUID

    from neo4j.graph import Relationship as Neo4jRelationship

    from infrahub.core.branch import Branch
    from infrahub.core.node import Node
    from infrahub.core.relationship import Relationship
    from infrahub.core.schema import NodeSchema, RelationshipSchema
    from infrahub.database import InfrahubDatabase

# pylint: disable=redefined-builtin,too-many-lines

log = get_logger()


@dataclass
class RelData:
    """Represent a relationship object in the database."""

    db_id: str
    branch: str
    type: str
    status: str

    @classmethod
    def from_db(cls, obj: Neo4jRelationship) -> RelData:
        return cls(db_id=obj.element_id, branch=obj.get("branch"), type=obj.type, status=obj.get("status"))


@dataclass
class FlagPropertyData:
    name: str
    prop_db_id: str
    rel: RelData
    value: bool


@dataclass
class NodePropertyData:
    name: str
    prop_db_id: str
    rel: RelData
    value: UUID


@dataclass
class RelationshipPeerData:
    branch: str

    source_id: UUID
    """UUID of the Source Node."""

    peer_id: UUID
    """UUID of the Peer Node."""

    peer_kind: str
    """Kind of the Peer Node."""

    properties: dict[str, FlagPropertyData | NodePropertyData]
    """UUID of the Relationship Node."""

    rel_node_id: UUID | None = None
    """UUID of the Relationship Node."""

    peer_db_id: str | None = None
    """Internal DB ID of the Peer Node."""

    rel_node_db_id: str | None = None
    """Internal DB ID of the Relationship Node."""

    rels: list[RelData] | None = None
    """Both relationships pointing at this Relationship Node."""

    updated_at: str | None = None

    def rel_ids_per_branch(self) -> dict[str, list[str | int]]:
        response = defaultdict(list)
        for rel in self.rels:
            response[rel.branch].append(rel.db_id)

        for prop in self.properties.values():
            response[prop.rel.branch].append(prop.rel.db_id)

        return response


@dataclass
class RelationshipPeersData:
    id: UUID
    identifier: str
    source_id: UUID
    source_kind: str
    destination_id: UUID
    destination_kind: str

    def reversed(self) -> RelationshipPeersData:
        return RelationshipPeersData(
            id=self.id,
            identifier=self.identifier,
            source_id=self.destination_id,
            source_kind=self.destination_kind,
            destination_id=self.source_id,
            destination_kind=self.source_kind,
        )


@dataclass
class FullRelationshipIdentifier:
    identifier: str
    source_kind: str
    destination_kind: str


class RelationshipQuery(Query):
    def __init__(
        self,
        rel: type[Relationship] | Relationship | None = None,
        rel_type: str | None = None,
        source: Node | None = None,
        source_id: UUID | None = None,
        destination: Node | None = None,
        destination_id: UUID | None = None,
        schema: RelationshipSchema | None = None,
        branch: Branch | None = None,
        at: Timestamp | str | None = None,
        **kwargs,
    ):
        if not source and not source_id:
            raise ValueError("Either source or source_id must be provided.")
        if not rel and not rel_type:
            raise ValueError("Either rel or rel_type must be provided.")
        if not inspect.isclass(rel) and not hasattr(rel, "schema"):
            raise ValueError("Rel must be a Relationship class or an instance of Relationship.")
        if not schema and inspect.isclass(rel) and not hasattr(rel, "schema"):
            raise ValueError("Either an instance of Relationship or a valid schema must be provided.")

        self.source_id = source_id or source.id
        self.source = source

        # Destination is optional because not all RelationshipQuery needs it
        # If a query must have a destination defined, the validation must be done in the query specific init
        self.destination = destination
        self.destination_id = destination_id
        if not self.destination_id and destination:
            self.destination_id = destination.id

        self.rel = rel
        self.rel_type = rel_type or self.rel.rel_type
        self.schema = schema or self.rel.schema

        if not branch and inspect.isclass(rel) and not hasattr(rel, "branch"):
            raise ValueError("Either an instance of Relationship or a valid branch must be provided.")

        self.branch = branch or self.rel.branch

        if at:
            self.at = Timestamp(at)
        elif inspect.isclass(rel) and hasattr(rel, "at"):
            self.at = self.rel.at
        else:
            self.at = Timestamp()

        super().__init__(**kwargs)

    def get_relationship_properties_dict(self, status: RelationshipStatus) -> dict[str, str | None]:
        rel_prop_dict = {
            "branch": self.branch.name,
            "branch_level": self.branch.hierarchy_level,
            "status": status.value,
            "from": self.at.to_string(),
        }
        if self.schema.hierarchical:
            rel_prop_dict["hierarchy"] = self.schema.hierarchical
        return rel_prop_dict


class RelationshipCreateQuery(RelationshipQuery):
    name = "relationship_create"

    type: QueryType = QueryType.WRITE

    def __init__(
        self,
        destination: Node = None,
        destination_id: UUID | None = None,
        **kwargs,
    ):
        if not destination and not destination_id:
            raise ValueError("Either destination or destination_id must be provided.")

        super().__init__(destination=destination, destination_id=destination_id, **kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:  # noqa: ARG002
        self.params["source_id"] = self.source_id
        self.params["destination_id"] = self.destination_id
        self.params["name"] = self.schema.identifier
        self.params["branch_support"] = self.schema.branch.value

        self.params["uuid"] = str(UUIDT())

        self.params["branch"] = self.branch.name
        self.params["branch_level"] = self.branch.hierarchy_level
        self.params["at"] = self.at.to_string()

        self.params["is_protected"] = self.rel.is_protected
        self.params["is_visible"] = self.rel.is_visible

        query_match = """
        MATCH (s:Node { uuid: $source_id })
        MATCH (d:Node { uuid: $destination_id })
        """
        self.add_to_query(query_match)

        self.query_add_all_node_property_match()

        self.params["rel_prop"] = self.get_relationship_properties_dict(status=RelationshipStatus.ACTIVE)
        arrows = self.schema.get_query_arrows()
        r1 = f"{arrows.left.start}[r1:{self.rel_type} $rel_prop ]{arrows.left.end}"
        r2 = f"{arrows.right.start}[r2:{self.rel_type} $rel_prop ]{arrows.right.end}"

        query_create = """
        CREATE (rl:Relationship { uuid: $uuid, name: $name, branch_support: $branch_support })
        CREATE (s)%s(rl)
        CREATE (rl)%s(d)
        MERGE (ip:Boolean { value: $is_protected })
        MERGE (iv:Boolean { value: $is_visible })
        CREATE (rl)-[r3:IS_PROTECTED $rel_prop ]->(ip)
        CREATE (rl)-[r4:IS_VISIBLE $rel_prop ]->(iv)
        """ % (
            r1,
            r2,
        )

        self.add_to_query(query_create)
        self.return_labels = ["s", "d", "rl", "r1", "r2", "r3", "r4"]
        self.query_add_all_node_property_create()

    def query_add_all_node_property_match(self) -> None:
        for prop_name in self.rel._node_properties:
            if hasattr(self.rel, f"{prop_name}_id") and getattr(self.rel, f"{prop_name}_id"):
                self.query_add_node_property_match(name=prop_name)

    def query_add_node_property_match(self, name: str) -> None:
        self.add_to_query("MATCH (%s { uuid: $prop_%s_id })" % (name, name))
        self.params[f"prop_{name}_id"] = getattr(self.rel, f"{name}_id")
        self.return_labels.append(name)

    def query_add_all_node_property_create(self) -> None:
        for prop_name in self.rel._node_properties:
            if hasattr(self.rel, f"{prop_name}_id") and getattr(self.rel, f"{prop_name}_id"):
                self.query_add_node_property_create(name=prop_name)

    def query_add_node_property_create(self, name: str) -> None:
        query = """
        CREATE (rl)-[:HAS_%s { branch: $branch, branch_level: $branch_level, status: "active", from: $at }]->(%s)
        """ % (
            name.upper(),
            name,
        )
        self.add_to_query(query)


class RelationshipUpdatePropertyQuery(RelationshipQuery):
    name = "relationship_property_update"
    type = QueryType.WRITE

    def __init__(
        self,
        properties_to_update: list[str],
        data: RelationshipPeerData,
        **kwargs,
    ):
        self.properties_to_update = properties_to_update
        self.data = data

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:  # noqa: ARG002
        self.params["rel_node_id"] = self.data.rel_node_id
        self.params["branch"] = self.branch.name
        self.params["branch_level"] = self.branch.hierarchy_level
        self.params["at"] = self.at.to_string()

        query = """
        MATCH (rl:Relationship { uuid: $rel_node_id })
        """
        self.add_to_query(query)

        self.query_add_all_flag_property_merge()
        self.query_add_all_node_property_merge()

        self.query_add_all_flag_property_create()
        self.query_add_all_node_property_create()

    def query_add_all_flag_property_merge(self) -> None:
        for prop_name in self.rel._flag_properties:
            if prop_name in self.properties_to_update:
                self.query_add_flag_property_merge(name=prop_name)

    def query_add_flag_property_merge(self, name: str) -> None:
        self.add_to_query("MERGE (prop_%s:Boolean { value: $prop_%s })" % (name, name))
        self.params[f"prop_{name}"] = getattr(self.rel, name)
        self.return_labels.append(f"prop_{name}")

    def query_add_all_node_property_merge(self) -> None:
        for prop_name in self.rel._node_properties:
            if prop_name in self.properties_to_update:
                self.query_add_node_property_merge(name=prop_name)

    def query_add_node_property_merge(self, name: str) -> None:
        self.add_to_query("MERGE (prop_%s:Node { uuid: $prop_%s })" % (name, name))
        self.params[f"prop_{name}"] = getattr(self.rel, f"{name}_id")
        self.return_labels.append(f"prop_{name}")

    def query_add_all_flag_property_create(self) -> None:
        for prop_name in self.rel._flag_properties:
            if prop_name in self.properties_to_update:
                self.query_add_flag_property_create(name=prop_name)

    def query_add_flag_property_create(self, name: str) -> None:
        query = """
        CREATE (rl)-[:%s { branch: $branch, branch_level: $branch_level, status: "active", from: $at }]->(prop_%s)
        """ % (
            name.upper(),
            name,
        )
        self.add_to_query(query)

    def query_add_all_node_property_create(self) -> None:
        for prop_name in self.rel._node_properties:
            if prop_name in self.properties_to_update:
                self.query_add_node_property_create(name=prop_name)

    def query_add_node_property_create(self, name: str) -> None:
        query = """
        CREATE (rl)-[:%s { branch: $branch, branch_level: $branch_level, status: "active", from: $at }]->(prop_%s)
        """ % (
            "HAS_" + name.upper(),
            name,
        )
        self.add_to_query(query)


class RelationshipDataDeleteQuery(RelationshipQuery):
    name = "relationship_data_delete"
    type = QueryType.WRITE

    def __init__(
        self,
        data: RelationshipPeerData,
        **kwargs,
    ):
        self.data = data
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:  # noqa: ARG002
        self.params["source_id"] = self.source_id
        self.params["destination_id"] = self.data.peer_id
        self.params["rel_node_id"] = self.data.rel_node_id
        self.params["name"] = self.schema.identifier
        self.params["branch"] = self.branch.name
        self.params["branch_level"] = self.branch.hierarchy_level
        self.params["at"] = self.at.to_string()

        # -----------------------------------------------------------------------
        # Match all nodes, including properties
        # -----------------------------------------------------------------------
        query = """
        MATCH (s:Node { uuid: $source_id })
        MATCH (d:Node { uuid: $destination_id })
        MATCH (rl:Relationship { uuid: $rel_node_id })
        """
        self.add_to_query(query)
        self.return_labels = ["s", "d", "rl"]

        for prop_name, prop in self.data.properties.items():
            self.add_to_query(
                "MATCH (prop_%(prop_name)s) WHERE %(id_func)s(prop_%(prop_name)s) = $prop_%(prop_name)s_id"
                % {"prop_name": prop_name, "id_func": db.get_id_function_name()}
            )
            self.params[f"prop_{prop_name}_id"] = db.to_database_id(prop.prop_db_id)
            self.return_labels.append(f"prop_{prop_name}")

        self.params["rel_prop"] = self.get_relationship_properties_dict(status=RelationshipStatus.DELETED)

        arrows = self.schema.get_query_arrows()
        r1 = f"{arrows.left.start}[r1:{self.rel_type} $rel_prop ]{arrows.left.end}"
        r2 = f"{arrows.right.start}[r2:{self.rel_type} $rel_prop ]{arrows.right.end}"

        # -----------------------------------------------------------------------
        # Create all the DELETE relationships, including properties
        # -----------------------------------------------------------------------
        query = """
        CREATE (s)%s(rl)
        CREATE (rl)%s(d)
        """ % (
            r1,
            r2,
        )
        self.add_to_query(query)
        self.return_labels.extend(["r1", "r2"])

        for prop_name, prop in self.data.properties.items():
            self.add_to_query(
                "CREATE (prop_%s)<-[rel_prop_%s:%s $rel_prop ]-(rl)" % (prop_name, prop_name, prop.rel.type),
            )
            self.return_labels.append(f"rel_prop_{prop_name}")


class RelationshipDeleteQuery(RelationshipQuery):
    name = "relationship_delete"
    type = QueryType.WRITE

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if inspect.isclass(self.rel):
            raise TypeError("An instance of Relationship must be provided to RelationshipDeleteQuery")

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:  # noqa: ARG002
        rel_filter, rel_params = self.branch.get_query_filter_path(at=self.at, variable_name="edge")
        self.params["source_id"] = self.source_id
        self.params["destination_id"] = self.destination_id
        self.params["rel_id"] = self.rel.id
        self.params["branch"] = self.branch.name
        self.params["rel_prop"] = self.get_relationship_properties_dict(status=RelationshipStatus.DELETED)
        self.params["at"] = self.at.to_string()
        self.params.update(rel_params)

        arrows = self.schema.get_query_arrows()
        r1 = f"{arrows.left.start}[r1:{self.rel_type} $rel_prop ]{arrows.left.end}"
        r2 = f"{arrows.right.start}[r2:{self.rel_type} $rel_prop ]{arrows.right.end}"

        query = """
        MATCH (s:Node { uuid: $source_id })-[:IS_RELATED]-(rl:Relationship {uuid: $rel_id})-[:IS_RELATED]-(d:Node { uuid: $destination_id })
        WITH s, rl, d
        LIMIT 1
        CREATE (s)%(r1)s(rl)
        CREATE (rl)%(r2)s(d)
        WITH rl
        CALL {
            WITH rl
            MATCH (rl)-[edge:IS_VISIBLE]->(visible)
            WHERE %(rel_filter)s AND edge.status = "active"
            WITH rl, edge, visible
            ORDER BY edge.branch_level DESC
            LIMIT 1
            CREATE (rl)-[deleted_edge:IS_VISIBLE $rel_prop]->(visible)
            WITH edge
            WHERE edge.branch = $branch
            SET edge.to = $at
        }
        CALL {
            WITH rl
            MATCH (rl)-[edge:IS_PROTECTED]->(protected)
            WHERE %(rel_filter)s AND edge.status = "active"
            WITH rl, edge, protected
            ORDER BY edge.branch_level DESC
            LIMIT 1
            CREATE (rl)-[deleted_edge:IS_PROTECTED $rel_prop]->(protected)
            WITH edge
            WHERE edge.branch = $branch
            SET edge.to = $at
        }
        CALL {
            WITH rl
            MATCH (rl)-[edge:HAS_OWNER]->(owner_node)
            WHERE %(rel_filter)s AND edge.status = "active"
            WITH rl, edge, owner_node
            ORDER BY edge.branch_level DESC
            LIMIT 1
            CREATE (rl)-[deleted_edge:HAS_OWNER $rel_prop]->(owner_node)
            WITH edge
            WHERE edge.branch = $branch
            SET edge.to = $at
        }
        CALL {
            WITH rl
            MATCH (rl)-[edge:HAS_SOURCE]->(source_node)
            WHERE %(rel_filter)s AND edge.status = "active"
            WITH rl, edge, source_node
            ORDER BY edge.branch_level DESC
            LIMIT 1
            CREATE (rl)-[deleted_edge:HAS_SOURCE $rel_prop]->(source_node)
            WITH edge
            WHERE edge.branch = $branch
            SET edge.to = $at
        }
        """ % {"r1": r1, "r2": r2, "rel_filter": rel_filter}

        self.params["at"] = self.at.to_string()
        self.return_labels = ["rl"]

        self.add_to_query(query)


class RelationshipGetPeerQuery(Query):
    name = "relationship_get_peer"
    type = QueryType.READ

    def __init__(
        self,
        filters: dict | None = None,
        source: Node | None = None,
        source_ids: list[str] | None = None,
        source_kind: str | None = None,
        rel: type[Relationship] | Relationship | None = None,
        rel_type: str | None = None,
        schema: RelationshipSchema | None = None,
        branch: Branch | None = None,
        at: Timestamp | str | None = None,
        **kwargs,
    ):
        if not source and not source_ids:
            raise ValueError("Either source or source_ids must be provided.")
        if not rel and not rel_type:
            raise ValueError("Either rel or rel_type must be provided.")
        if rel and not inspect.isclass(rel) and not hasattr(rel, "schema"):
            raise ValueError("Rel must be a Relationship class or an instance of Relationship.")
        if not schema and inspect.isclass(rel) and not hasattr(rel, "schema"):
            raise ValueError("Either an instance of Relationship or a valid schema must be provided.")

        self.filters = filters or {}
        self.source_ids = source_ids or [source.id]
        self.source = source

        self.source_kind = source_kind or "Node"
        if source and not source_kind:
            self.source_kind = source.get_kind()

        self.rel = rel
        self.rel_type = rel_type or self.rel.rel_type
        self.schema = schema or self.rel.schema

        if not branch and inspect.isclass(rel) and not hasattr(rel, "branch"):
            raise ValueError("Either an instance of Relationship or a valid branch must be provided.")

        self.branch = branch or self.rel.branch

        if not at and inspect.isclass(rel) and hasattr(rel, "at"):
            self.at = self.rel.at
        else:
            self.at = Timestamp(at)

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:  # noqa: ARG002
        branch_filter, branch_params = self.branch.get_query_filter_path(
            at=self.at, branch_agnostic=self.branch_agnostic
        )
        self.params.update(branch_params)
        self.order_by = []

        peer_schema = self.schema.get_peer_schema(db=db, branch=self.branch)

        self.params["source_ids"] = self.source_ids
        self.params["rel_identifier"] = self.schema.identifier
        self.params["peer_kind"] = self.schema.peer
        self.params["source_kind"] = self.source_kind

        arrows = self.schema.get_query_arrows()

        path_str = (
            f"{arrows.left.start}[:IS_RELATED]{arrows.left.end}(rl){arrows.right.start}[:IS_RELATED]{arrows.right.end}"
        )

        branch_level_str = "reduce(br_lvl = 0, r in relationships(path) | br_lvl + r.branch_level)"
        froms_str = db.render_list_comprehension(items="relationships(path)", item_name="from")
        query = """
        MATCH (source_node:Node)%(arrow_left_start)s[:IS_RELATED]%(arrow_left_end)s(rl:Relationship { name: $rel_identifier })
        WHERE source_node.uuid IN $source_ids
        WITH DISTINCT source_node, rl
        CALL {
            WITH rl, source_node
            MATCH path = (source_node)%(path)s(peer:Node)
            WHERE
                $source_kind IN LABELS(source_node) AND
                peer.uuid <> source_node.uuid AND
                $peer_kind IN LABELS(peer) AND
                all(r IN relationships(path) WHERE (%(branch_filter)s))
            WITH source_node, peer, rl, relationships(path) as rels, %(branch_level)s AS branch_level, %(froms)s AS froms
            RETURN peer as peer, rels, rl as rl1
            ORDER BY branch_level DESC, froms[-1] DESC, froms[-2] DESC
            LIMIT 1
        }
        WITH peer, rl1 as rl, rels, source_node
        """ % {
            "path": path_str,
            "branch_filter": branch_filter,
            "branch_level": branch_level_str,
            "froms": froms_str,
            "arrow_left_start": arrows.left.start,
            "arrow_left_end": arrows.left.end,
        }

        self.add_to_query(query)
        where_clause = ['all(r IN rels WHERE r.status = "active")']
        clean_filters = extract_field_filters(field_name=self.schema.name, filters=self.filters)

        if clean_filters and "id" in clean_filters or "ids" in clean_filters:
            where_clause.append("peer.uuid IN $peer_ids")
            self.params["peer_ids"] = clean_filters.get("ids", [])
            if clean_filters.get("id", None):
                self.params["peer_ids"].append(clean_filters.get("id"))

        self.add_to_query("WHERE " + " AND ".join(where_clause))

        self.return_labels = ["rl", "peer", "rels", "source_node"]

        # ----------------------------------------------------------------------------
        # FILTER Results
        # ----------------------------------------------------------------------------
        filter_cnt = 0
        for peer_filter_name, peer_filter_value in clean_filters.items():
            if "__" not in peer_filter_name:
                continue

            filter_cnt += 1

            filter_field_name, filter_next_name = peer_filter_name.split("__", maxsplit=1)

            if filter_field_name not in peer_schema.valid_input_names:
                continue

            field = peer_schema.get_field(filter_field_name)

            subquery, subquery_params, subquery_result_name = await build_subquery_filter(
                db=db,
                node_alias="peer",
                field=field,
                name=filter_field_name,
                filter_name=filter_next_name,
                filter_value=peer_filter_value,
                branch_filter=branch_filter,
                branch=self.branch,
                subquery_idx=filter_cnt,
            )
            self.params.update(subquery_params)

            with_str = ", ".join(
                [f"{subquery_result_name} as {label}" if label == "peer" else label for label in self.return_labels]
            )
            self.add_subquery(subquery=subquery, with_clause=with_str)

        # ----------------------------------------------------------------------------
        # QUERY Properties
        # ----------------------------------------------------------------------------
        query = """
        CALL {
            WITH rl
            MATCH (rl)-[r:IS_VISIBLE]-(is_visible)
            WHERE %(branch_filter)s
            RETURN r AS rel_is_visible, is_visible
            ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
            LIMIT 1
        }
        CALL {
            WITH rl
            MATCH (rl)-[r:IS_PROTECTED]-(is_protected)
            WHERE %(branch_filter)s
            RETURN r AS rel_is_protected, is_protected
            ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
            LIMIT 1
        }
        """ % {"branch_filter": branch_filter}

        self.add_to_query(query)

        self.update_return_labels(["rel_is_visible", "rel_is_protected", "is_visible", "is_protected"])

        # Add Node Properties
        # We must query them one by one otherwise the second one won't return
        for node_prop in ["source", "owner"]:
            query = """
        CALL {
            WITH rl
            OPTIONAL MATCH (rl)-[r:HAS_%(node_prop_type)s]-(%(node_prop)s)
            WHERE %(branch_filter)s
            RETURN r AS rel_%(node_prop)s, %(node_prop)s
            ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
            LIMIT 1
        }
            """ % {
                "node_prop": node_prop,
                "node_prop_type": node_prop.upper(),
                "branch_filter": branch_filter,
            }
            self.add_to_query(query)
            self.update_return_labels([f"rel_{node_prop}", node_prop])

        self.add_to_query("WITH " + ",".join(self.return_labels))

        # ----------------------------------------------------------------------------
        # ORDER Results
        # ----------------------------------------------------------------------------
        if hasattr(peer_schema, "order_by") and peer_schema.order_by:
            order_cnt = 1

            for order_by_value in peer_schema.order_by:
                order_by_field_name, order_by_next_name = order_by_value.split("__", maxsplit=1)

                field = peer_schema.get_field(order_by_field_name)

                subquery, subquery_params, subquery_result_name = await build_subquery_order(
                    db=db,
                    field=field,
                    node_alias="peer",
                    name=order_by_field_name,
                    order_by=order_by_next_name,
                    branch_filter=branch_filter,
                    branch=self.branch,
                    subquery_idx=order_cnt,
                )
                self.order_by.append(subquery_result_name)
                self.params.update(subquery_params)

                self.add_subquery(subquery=subquery)

                order_cnt += 1

        else:
            self.order_by.append("peer.uuid")

    def get_peer_ids(self) -> list[str]:
        """Return a list of UUID of nodes associated with this relationship."""

        return [peer.peer_id for peer in self.get_peers()]

    def get_peers(self) -> Generator[RelationshipPeerData, None, None]:
        for result in self.get_results_group_by(("peer", "uuid"), ("source_node", "uuid")):
            rels = result.get("rels")
            data = RelationshipPeerData(
                source_id=result.get_node("source_node").get("uuid"),
                peer_id=result.get_node("peer").get("uuid"),
                peer_kind=result.get_node("peer").get("kind"),
                rel_node_db_id=result.get("rl").element_id,
                rel_node_id=result.get("rl").get("uuid"),
                updated_at=rels[0]["from"],
                rels=[RelData.from_db(rel) for rel in rels],
                branch=self.branch.name,
                properties={},
            )

            if hasattr(self.rel, "_flag_properties"):
                for prop in self.rel._flag_properties:
                    if prop_node := result.get(prop):
                        data.properties[prop] = FlagPropertyData(
                            name=prop,
                            prop_db_id=prop_node.element_id,
                            rel=RelData.from_db(result.get(f"rel_{prop}")),
                            value=prop_node.get("value"),
                        )

            if hasattr(self.rel, "_node_properties"):
                for prop in self.rel._node_properties:
                    if prop_node := result.get(prop):
                        data.properties[prop] = NodePropertyData(
                            name=prop,
                            prop_db_id=prop_node.element_id,
                            rel=RelData.from_db(result.get(f"rel_{prop}")),
                            value=prop_node.get("uuid"),
                        )

            yield data


class RelationshipGetQuery(RelationshipQuery):
    name = "relationship_get"

    type: QueryType = QueryType.READ

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:  # noqa: ARG002
        self.params["source_id"] = self.source_id
        self.params["destination_id"] = self.destination_id
        self.params["name"] = self.schema.identifier
        self.params["branch"] = self.branch.name

        rels_filter, rels_params = self.branch.get_query_filter_relationships(
            rel_labels=["r1", "r2"], at=self.at.to_string(), include_outside_parentheses=True
        )

        self.params.update(rels_params)

        arrows = self.schema.get_query_arrows()
        r1 = f"{arrows.left.start}[r1:{self.rel.rel_type}]{arrows.left.end}"
        r2 = f"{arrows.right.start}[r2:{self.rel.rel_type}]{arrows.right.end}"

        query = """
        MATCH (s:Node { uuid: $source_id })
        MATCH (d:Node { uuid: $destination_id })
        MATCH (s)%s(rl:Relationship { name: $name })%s(d)
        WHERE %s
        """ % (
            r1,
            r2,
            "\n AND ".join(rels_filter),
        )

        self.params["at"] = self.at.to_string()

        self.add_to_query(query)
        self.return_labels = ["s", "d", "rl", "r1", "r2"]


class RelationshipGetByIdentifierQuery(Query):
    name = "relationship_get_identifier"
    type = QueryType.READ

    def __init__(
        self,
        identifiers: list[str] | None = None,
        full_identifiers: list[FullRelationshipIdentifier] | None = None,
        excluded_namespaces: list[str] | None = None,
        **kwargs,
    ) -> None:
        if (not identifiers and not full_identifiers) or (identifiers and full_identifiers):
            raise ValueError("one and only one of identifiers or full_identifiers is required")

        if full_identifiers:
            self.identifiers = list({i.identifier for i in full_identifiers})
            self.full_identifiers = full_identifiers
        else:
            self.identifiers = identifiers
            self.full_identifiers = []
        self.excluded_namespaces = excluded_namespaces or []

        # Always exclude relationships with internal nodes
        if "Internal" not in self.excluded_namespaces:
            self.excluded_namespaces.append("Internal")

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:  # noqa: ARG002
        self.params["identifiers"] = self.identifiers
        self.params["full_identifiers"] = [
            [full_id.source_kind, full_id.identifier, full_id.destination_kind] for full_id in self.full_identifiers
        ]
        self.params["excluded_namespaces"] = self.excluded_namespaces
        self.params["branch"] = self.branch.name
        self.params["at"] = self.at.to_string()

        rels_filter, rels_params = self.branch.get_query_filter_relationships(
            rel_labels=["r1", "r2"], at=self.at.to_string(), include_outside_parentheses=True
        )
        self.params.update(rels_params)

        query = """
        MATCH (rl:Relationship)
        WHERE rl.name IN $identifiers
        CALL {
            WITH rl
            MATCH (src:Node)-[r1:IS_RELATED]-(rl:Relationship)-[r2:IS_RELATED]-(dst:Node)
            WHERE (size($full_identifiers) = 0 OR [src.kind, rl.name, dst.kind] in $full_identifiers)
            AND NOT src.namespace IN $excluded_namespaces
            AND NOT dst.namespace IN $excluded_namespaces
            AND %s
            RETURN src, dst, r1, r2, rl as rl1
            ORDER BY r1.branch_level DESC, r2.branch_level DESC, r1.from DESC, r2.from DESC
            LIMIT 1
        }
        WITH src, dst, r1, r2, rl1 as rl
        WHERE r1.status = "active" AND r2.status = "active"
        """ % ("\n AND ".join(rels_filter),)

        self.add_to_query(query)
        self.return_labels = ["src", "dst", "rl"]

    def get_peers(self) -> Generator[RelationshipPeersData, None, None]:
        for result in self.get_results():
            data = RelationshipPeersData(
                id=result.get("rl").get("uuid"),
                identifier=result.get("rl").get("name"),
                source_id=result.get("src").get("uuid"),
                source_kind=result.get("src").get("kind"),
                destination_id=result.get("dst").get("uuid"),
                destination_kind=result.get("dst").get("kind"),
            )
            yield data


class RelationshipCountPerNodeQuery(Query):
    name = "relationship_count_per_node"
    type: QueryType = QueryType.READ

    def __init__(
        self,
        node_ids: list[str],
        identifier: str,
        direction: RelationshipDirection,
        **kwargs,
    ):
        self.node_ids = node_ids
        self.identifier = identifier
        self.direction = direction

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:  # noqa: ARG002
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string())
        self.params.update(branch_params)

        self.params["peer_ids"] = self.node_ids
        self.params["rel_identifier"] = self.identifier

        path = "-[r:IS_RELATED]-"
        if self.direction == RelationshipDirection.OUTBOUND:
            path = "-[r:IS_RELATED]->"
        elif self.direction == RelationshipDirection.INBOUND:
            path = "<-[r:IS_RELATED]-"

        query = """
        MATCH (peer_node:Node)%(path)s(rl:Relationship { name: $rel_identifier })
        WHERE peer_node.uuid IN $peer_ids AND %(branch_filter)s
        CALL {
            WITH rl
            MATCH path = (peer_node:Node)%(path)s(rl)
            WHERE peer_node.uuid IN $peer_ids AND %(branch_filter)s
            RETURN peer_node as peer, r as r1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH peer as peer_node, r1 as r
        WHERE r.status = "active"
        """ % {"branch_filter": branch_filter, "path": path}

        self.add_to_query(query)
        self.return_labels = ["peer_node.uuid", "COUNT(peer_node.uuid) as nbr_peers"]

    async def get_count_per_peer(self) -> dict[str, int]:
        data: dict[str, int] = {}
        for result in self.results:
            data[result.get("peer_node.uuid")] = result.get("nbr_peers")

        for node_id in self.node_ids:
            if node_id not in data:
                data[node_id] = 0

        return data


class RelationshipDeleteAllQuery(Query):
    """
    Delete all relationships linked to a given node on a given branch at a given time. For every IS_RELATED edge:
    - Set `to` time if an active edge exist on the same branch.
    - Create `deleted` edge.
    - Apply above to every edges linked to any connected Relationship node.
    This query returns node uuids/kinds and corresponding relationship identifiers of deleted nodes,
    that are later used to update node changelog.
    """

    name = "node_delete_all_relationships"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, node_id: str, **kwargs):
        self.node_id = node_id
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:  # noqa: ARG002
        self.params["source_id"] = kwargs["node_id"]
        self.params["branch"] = self.branch.name

        self.params["rel_prop"] = {
            "branch": self.branch.name,
            "branch_level": self.branch.hierarchy_level,
            "status": RelationshipStatus.DELETED.value,
            "from": self.at.to_string(),
        }

        self.params["at"] = self.at.to_string()

        active_rel_filter, rel_params = self.branch.get_query_filter_path(
            at=self.at, variable_name="active_edge", branch_agnostic=self.branch_agnostic
        )
        self.params.update(rel_params)

        query = """
        MATCH (s:Node { uuid: $source_id })-[active_edge:IS_RELATED]-(rl:Relationship)
        WHERE %(active_rel_filter)s AND active_edge.status = "active"
        WITH DISTINCT rl
        """ % {"active_rel_filter": active_rel_filter}

        edge_types = [
            DatabaseEdgeType.IS_VISIBLE.value,
            DatabaseEdgeType.IS_PROTECTED.value,
            DatabaseEdgeType.HAS_OWNER.value,
            DatabaseEdgeType.HAS_SOURCE.value,
        ]

        for arrow_left, arrow_right in (("<-", "-"), ("-", "->")):
            for edge_type in edge_types:
                sub_query = """
                    CALL {
                        WITH rl
                        MATCH (rl)%(arrow_left)s[active_edge:%(edge_type)s]%(arrow_right)s(n)
                        WHERE %(active_rel_filter)s AND active_edge.status ="active"
                        CREATE (rl)%(arrow_left)s[deleted_edge:%(edge_type)s $rel_prop]%(arrow_right)s(n)
                        SET deleted_edge.hierarchy = active_edge.hierarchy
                        WITH active_edge, n
                        WHERE active_edge.branch = $branch AND active_edge.to IS NULL
                        SET active_edge.to = $at
                    }
                """ % {
                    "arrow_left": arrow_left,
                    "arrow_right": arrow_right,
                    "active_rel_filter": active_rel_filter,
                    "edge_type": edge_type,
                }

                query += sub_query

        # We only want to return uuid/kind of `Node` connected through `IS_RELATED` edges.
        query += """
        CALL {
            WITH rl
            MATCH (rl)-[active_edge:IS_RELATED]->(n)
            WHERE %(active_rel_filter)s AND active_edge.status ="active"
            CREATE (rl)-[deleted_edge:IS_RELATED $rel_prop]->(n)
            SET deleted_edge.hierarchy = active_edge.hierarchy
            WITH rl, active_edge, n
            WHERE active_edge.branch = $branch AND active_edge.to IS NULL
            SET active_edge.to = $at
            RETURN
                n.uuid as uuid,
                n.kind as kind,
                rl.name as rel_identifier,
                "outbound" as rel_direction

            UNION

            WITH rl
            MATCH (rl)<-[active_edge:IS_RELATED]-(n)
            WHERE %(active_rel_filter)s AND active_edge.status ="active"
            CREATE (rl)<-[deleted_edge:IS_RELATED $rel_prop]-(n)
            SET deleted_edge.hierarchy = active_edge.hierarchy
            WITH rl, active_edge, n
            WHERE active_edge.branch = $branch AND active_edge.to IS NULL
            SET active_edge.to = $at
            RETURN
                n.uuid as uuid,
                n.kind as kind,
                rl.name as rel_identifier,
                "inbound" as rel_direction
        }
        RETURN DISTINCT uuid, kind, rel_identifier, rel_direction
        """ % {
            "active_rel_filter": active_rel_filter,
        }

        self.add_to_query(query)

    def get_deleted_relationships_changelog(
        self, node_schema: NodeSchema
    ) -> list[RelationshipCardinalityOneChangelog | RelationshipCardinalityManyChangelog]:
        rel_identifier_to_changelog_mapper = {}

        for result in self.get_results():
            peer_uuid = result.data["uuid"]
            if peer_uuid == self.node_id:
                continue

            rel_identifier = result.data["rel_identifier"]
            kind = result.data["kind"]
            deleted_rel_schemas = [
                rel_schema for rel_schema in node_schema.relationships if rel_schema.identifier == rel_identifier
            ]

            if len(deleted_rel_schemas) == 0:
                continue  # TODO Unidirectional relationship changelog should be handled, cf IFC-1319.

            if len(deleted_rel_schemas) > 2:
                log.error(f"Duplicated relationship schema with identifier {rel_identifier}")
                continue

            if len(deleted_rel_schemas) == 2:
                # Hierarchical schema nodes have 2 relationships with `parent_child` identifiers,
                # which are differentiated by their direction within the database.
                # assert rel_identifier != PARENT_CHILD_IDENTIFIER

                rel_direction = result.data["rel_direction"]
                deleted_rel_schema = (
                    deleted_rel_schemas[0]
                    if deleted_rel_schemas[0].direction.value == rel_direction
                    else deleted_rel_schemas[1]
                )
            else:
                deleted_rel_schema = deleted_rel_schemas[0]

            try:
                changelog_mapper = rel_identifier_to_changelog_mapper[rel_identifier]
            except KeyError:
                changelog_mapper = ChangelogRelationshipMapper(schema=deleted_rel_schema)
                rel_identifier_to_changelog_mapper[rel_identifier] = changelog_mapper

            changelog_mapper.delete_relationship(peer_id=peer_uuid, peer_kind=kind, rel_schema=deleted_rel_schema)

        return [changelog_mapper.changelog for changelog_mapper in rel_identifier_to_changelog_mapper.values()]



from __future__ import annotations

from typing import Sequence

from pydantic import Field

from ..shared import MigrationQuery, SchemaMigration


class PlaceholderDummyMigration(SchemaMigration):
    name: str = "dummy.placeholder"
    queries: Sequence[type[MigrationQuery]] = Field(default_factory=list)



from __future__ import annotations

import os

import pytest
from infrahub_testcontainers.helpers import TestInfrahubDocker
from packaging.version import InvalidVersion, Version

from .. import Config, InfrahubClient, InfrahubClientSync

INFRAHUB_VERSION = os.getenv("INFRAHUB_TESTING_IMAGE_VER", "latest")


def skip_version(min_infrahub_version: str | None = None, max_infrahub_version: str | None = None) -> bool:
    """
    Check if a test should be skipped depending on infrahub version.
    """
    try:
        version = Version(INFRAHUB_VERSION)
    except InvalidVersion:
        # We would typically end up here for development purpose while running a CI test against
        # unreleased versions of infrahub, like `stable` or `develop` branch.
        # For now, we consider this means we are testing against the most recent version of infrahub,
        # so we skip if the test should not be ran against a maximum version.
        return max_infrahub_version is None

    if min_infrahub_version is not None and version < Version(min_infrahub_version):
        return True

    return max_infrahub_version is not None and version > Version(max_infrahub_version)


class TestInfrahubDockerClient(TestInfrahubDocker):
    @pytest.fixture(scope="class")
    def infrahub_version(self) -> str:
        return INFRAHUB_VERSION

    @pytest.fixture(scope="class")
    def client(self, infrahub_port: int) -> InfrahubClient:
        return InfrahubClient(
            config=Config(username="admin", password="infrahub", address=f"http://localhost:{infrahub_port}")  # noqa: S106
        )

    @pytest.fixture(scope="class")
    def client_sync(self, infrahub_port: int) -> InfrahubClientSync:
        return InfrahubClientSync(
            config=Config(username="admin", password="infrahub", address=f"http://localhost:{infrahub_port}")  # noqa: S106
        )



from typing import TYPE_CHECKING

from infrahub.core import registry
from infrahub.core.constants import InfrahubKind
from infrahub.core.manager import NodeManager
from infrahub.database import InfrahubDatabase

from .models import RepositoryBranchInfo, RepositoryData

if TYPE_CHECKING:
    from infrahub.core.protocols import CoreGenericRepository


async def get_repositories_commit_per_branch(
    db: InfrahubDatabase,
) -> dict[str, RepositoryData]:
    """Get a list of all repositories and their commit on each branches.

    This method is similar to 'get_list_repositories' method in the Python SDK.

    NOTE: At some point, we should refactor this function to use a single Database query instead of one per branch
    """

    repositories: dict[str, RepositoryData] = {}

    for branch in list(registry.branch.values()):
        repos: list[CoreGenericRepository] = await NodeManager.query(
            db=db,
            branch=branch,
            fields={"id": None, "name": None, "commit": None, "internal_status": None},
            schema=InfrahubKind.GENERICREPOSITORY,
        )

        for repository in repos:
            repo_name = repository.name.value
            if repo_name not in repositories:
                repositories[repo_name] = RepositoryData(
                    repository_id=repository.get_id(),
                    repository_name=repo_name,
                    branches={},
                )

            repositories[repo_name].branches[branch.name] = repository.commit.value  # type: ignore[attr-defined]
            repositories[repo_name].branch_info[branch.name] = RepositoryBranchInfo(
                internal_status=repository.internal_status.value
            )

    return repositories



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol, runtime_checkable

if TYPE_CHECKING:
    import ipaddress

    from .context import RequestContext
    from .schema import MainSchemaTypes


@runtime_checkable
class RelatedNode(Protocol): ...


@runtime_checkable
class RelatedNodeSync(Protocol): ...


@runtime_checkable
class Attribute(Protocol):
    name: str
    id: str | None
    is_default: bool | None
    is_from_profile: bool | None
    is_inherited: bool | None
    updated_at: str | None
    is_visible: bool | None
    is_protected: bool | None


class String(Attribute):
    value: str


class StringOptional(Attribute):
    value: str | None


class DateTime(Attribute):
    value: str


class DateTimeOptional(Attribute):
    value: str | None


class HashedPassword(Attribute):
    value: str


class HashedPasswordOptional(Attribute):
    value: Any


class URL(Attribute):
    value: str


class URLOptional(Attribute):
    value: str | None


class MacAddress(Attribute):
    value: str


class MacAddressOptional(Attribute):
    value: str | None


class Dropdown(Attribute):
    value: str


class DropdownOptional(Attribute):
    value: str | None


class Enum(Attribute):
    value: str


class EnumOptional(Attribute):
    value: str | None


class Integer(Attribute):
    value: int


class IntegerOptional(Attribute):
    value: int | None


class IPHost(Attribute):
    value: ipaddress.IPv4Address | ipaddress.IPv6Address


class IPHostOptional(Attribute):
    value: ipaddress.IPv4Address | ipaddress.IPv6Address | None


class IPNetwork(Attribute):
    value: ipaddress.IPv4Network | ipaddress.IPv6Network


class IPNetworkOptional(Attribute):
    value: ipaddress.IPv4Network | ipaddress.IPv6Network | None


class Boolean(Attribute):
    value: bool


class BooleanOptional(Attribute):
    value: bool | None


class ListAttribute(Attribute):
    value: list[Any]


class ListAttributeOptional(Attribute):
    value: list[Any] | None


class JSONAttribute(Attribute):
    value: Any


class JSONAttributeOptional(Attribute):
    value: Any | None


class AnyAttribute(Attribute):
    value: float


class AnyAttributeOptional(Attribute):
    value: float | None


@runtime_checkable
class CoreNodeBase(Protocol):
    _schema: MainSchemaTypes
    _internal_id: str
    id: str  # NOTE this is incorrect, should be str | None
    display_label: str | None

    @property
    def hfid(self) -> list[str] | None: ...

    @property
    def hfid_str(self) -> str | None: ...

    def get_human_friendly_id(self) -> list[str] | None: ...

    def get_human_friendly_id_as_string(self, include_kind: bool = False) -> str | None: ...

    def get_kind(self) -> str: ...

    def get_all_kinds(self) -> list[str]: ...

    def get_branch(self) -> str: ...

    def is_ip_prefix(self) -> bool: ...

    def is_ip_address(self) -> bool: ...

    def is_resource_pool(self) -> bool: ...

    def get_raw_graphql_data(self) -> dict | None: ...

    def extract(self, params: dict[str, str]) -> dict[str, Any]: ...


@runtime_checkable
class CoreNode(CoreNodeBase, Protocol):
    async def save(
        self,
        allow_upsert: bool = False,
        update_group_context: bool | None = None,
        timeout: int | None = None,
        request_context: RequestContext | None = None,
    ) -> None: ...

    async def delete(self, timeout: int | None = None, request_context: RequestContext | None = None) -> None: ...

    async def update(
        self, do_full_update: bool, timeout: int | None = None, request_context: RequestContext | None = None
    ) -> None: ...

    async def create(
        self, allow_upsert: bool = False, timeout: int | None = None, request_context: RequestContext | None = None
    ) -> None: ...

    async def add_relationships(self, relation_to_update: str, related_nodes: list[str]) -> None: ...

    async def remove_relationships(self, relation_to_update: str, related_nodes: list[str]) -> None: ...


@runtime_checkable
class CoreNodeSync(CoreNodeBase, Protocol):
    def save(
        self,
        allow_upsert: bool = False,
        update_group_context: bool | None = None,
        timeout: int | None = None,
        request_context: RequestContext | None = None,
    ) -> None: ...

    def delete(self, timeout: int | None = None, request_context: RequestContext | None = None) -> None: ...

    def update(
        self, do_full_update: bool, timeout: int | None = None, request_context: RequestContext | None = None
    ) -> None: ...

    def create(
        self, allow_upsert: bool = False, timeout: int | None = None, request_context: RequestContext | None = None
    ) -> None: ...

    def add_relationships(self, relation_to_update: str, related_nodes: list[str]) -> None: ...

    def remove_relationships(self, relation_to_update: str, related_nodes: list[str]) -> None: ...



from infrahub.core.constants import (
    BranchSupportType,
    GeneratorInstanceStatus,
    InfrahubKind,
)
from infrahub.core.constants import RelationshipCardinality as Cardinality
from infrahub.core.constants import RelationshipKind as RelKind

from ...attribute_schema import AttributeSchema as Attr
from ...node_schema import NodeSchema
from ...relationship_schema import (
    RelationshipSchema as Rel,
)

core_generator_definition = NodeSchema(
    name="GeneratorDefinition",
    namespace="Core",
    include_in_menu=False,
    icon="mdi:state-machine",
    label="Generator Definition",
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["name__value"],
    branch=BranchSupportType.AWARE,
    uniqueness_constraints=[["name__value"]],
    generate_profile=False,
    inherit_from=[InfrahubKind.TASKTARGET],
    documentation="/topics/generator",
    attributes=[
        Attr(name="name", kind="Text", unique=True),
        Attr(name="description", kind="Text", optional=True),
        Attr(name="parameters", kind="JSON"),
        Attr(name="file_path", kind="Text"),
        Attr(name="class_name", kind="Text"),
        Attr(name="convert_query_response", kind="Boolean", optional=True, default_value=False),
    ],
    relationships=[
        Rel(
            name="query",
            peer=InfrahubKind.GRAPHQLQUERY,
            identifier="generator_definition__graphql_query",
            kind=RelKind.ATTRIBUTE,
            cardinality=Cardinality.ONE,
            optional=False,
        ),
        Rel(
            name="repository",
            peer=InfrahubKind.GENERICREPOSITORY,
            kind=RelKind.ATTRIBUTE,
            cardinality=Cardinality.ONE,
            identifier="generator_definition__repository",
            optional=False,
        ),
        Rel(
            name="targets",
            peer=InfrahubKind.GENERICGROUP,
            kind=RelKind.ATTRIBUTE,
            identifier="generator_definition___group",
            cardinality=Cardinality.ONE,
            optional=False,
        ),
    ],
)

core_generator_instance = NodeSchema(
    name="GeneratorInstance",
    namespace="Core",
    label="Generator Instance",
    include_in_menu=False,
    icon="mdi:file-document-outline",
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["name__value"],
    branch=BranchSupportType.LOCAL,
    generate_profile=False,
    inherit_from=[InfrahubKind.TASKTARGET],
    documentation="/topics/generator",
    attributes=[
        Attr(name="name", kind="Text"),
        Attr(name="status", kind="Text", enum=GeneratorInstanceStatus.available_types()),
    ],
    relationships=[
        Rel(
            name="object",
            peer=InfrahubKind.NODE,
            kind=RelKind.ATTRIBUTE,
            identifier="generator__node",
            cardinality=Cardinality.ONE,
            optional=False,
        ),
        Rel(
            name="definition",
            peer=InfrahubKind.GENERATORDEFINITION,
            kind=RelKind.ATTRIBUTE,
            identifier="generator__generator_definition",
            cardinality=Cardinality.ONE,
            optional=False,
        ),
    ],
)



from collections import defaultdict
from typing import AsyncGenerator, Generator, Iterable

from neo4j.exceptions import TransientError

from infrahub import config
from infrahub.core import registry
from infrahub.core.diff.query.field_summary import EnrichedDiffNodeFieldSummaryQuery
from infrahub.core.diff.query.summary_counts_enricher import (
    DiffFieldsSummaryCountsEnricherQuery,
    DiffNodesSummaryCountsEnricherQuery,
)
from infrahub.core.query.diff import DiffCountChanges
from infrahub.core.timestamp import Timestamp
from infrahub.database import InfrahubDatabase, retry_db_transaction
from infrahub.exceptions import ResourceNotFoundError
from infrahub.log import get_logger

from ..model.path import (
    ConflictSelection,
    EnrichedDiffConflict,
    EnrichedDiffNode,
    EnrichedDiffRoot,
    EnrichedDiffRootMetadata,
    EnrichedDiffs,
    EnrichedDiffsMetadata,
    EnrichedNodeCreateRequest,
    NodeDiffFieldSummary,
    TimeRange,
    TrackingId,
)
from ..query.all_conflicts import EnrichedDiffAllConflictsQuery
from ..query.delete_query import EnrichedDiffDeleteQuery
from ..query.diff_get import EnrichedDiffGetQuery
from ..query.diff_summary import DiffSummaryCounters, DiffSummaryQuery
from ..query.field_specifiers import EnrichedDiffFieldSpecifiersQuery
from ..query.filters import EnrichedDiffQueryFilters
from ..query.get_conflict_query import EnrichedDiffConflictQuery
from ..query.has_conflicts_query import EnrichedDiffHasConflictQuery
from ..query.merge_tracking_id import EnrichedDiffMergedTrackingIdQuery
from ..query.roots_metadata import EnrichedDiffRootsMetadataQuery
from ..query.save import EnrichedDiffRootsUpsertQuery, EnrichedNodeBatchCreateQuery, EnrichedNodesLinkQuery
from ..query.time_range_query import EnrichedDiffTimeRangeQuery
from ..query.update_conflict_query import EnrichedDiffConflictUpdateQuery
from .deserializer import EnrichedDiffDeserializer

log = get_logger()


class DiffRepository:
    def __init__(self, db: InfrahubDatabase, deserializer: EnrichedDiffDeserializer, max_save_batch_size: int = 1000):
        self.db = db
        self.deserializer = deserializer
        self.max_save_batch_size = max_save_batch_size

    async def _run_get_diff_query(
        self,
        base_branch_name: str,
        diff_branch_names: list[str],
        batch_size_limit: int,
        limit: int | None = None,
        from_time: Timestamp | None = None,
        to_time: Timestamp | None = None,
        filters: EnrichedDiffQueryFilters | None = None,
        offset: int = 0,
        include_parents: bool = True,
        max_depth: int | None = None,
        tracking_id: TrackingId | None = None,
        diff_ids: list[str] | None = None,
    ) -> list[EnrichedDiffRoot]:
        self.deserializer.initialize()
        final_row_number = None
        if limit:
            final_row_number = offset + limit
        has_more_data = True
        while has_more_data and (final_row_number is None or offset < final_row_number):
            if final_row_number is not None and offset + batch_size_limit > final_row_number:
                batch_size_limit = final_row_number - offset
            get_query = await EnrichedDiffGetQuery.init(
                db=self.db,
                base_branch_name=base_branch_name,
                diff_branch_names=diff_branch_names,
                from_time=from_time,
                to_time=to_time,
                filters=filters,
                max_depth=max_depth,
                limit=batch_size_limit,
                offset=offset,
                tracking_id=tracking_id,
                diff_ids=diff_ids,
            )
            log.info(f"Beginning enriched diff get query {batch_size_limit=}, {offset=}")
            await get_query.execute(db=self.db)
            log.info("Enriched diff get query complete")
            last_result = None
            for query_result in get_query.get_results():
                await self.deserializer.read_result(result=query_result, include_parents=include_parents)
                last_result = query_result
            has_more_data = False
            if last_result:
                has_more_data = last_result.get_as_type("has_more_data", bool)
            offset += batch_size_limit
        return await self.deserializer.deserialize()

    async def get(
        self,
        base_branch_name: str,
        diff_branch_names: list[str],
        from_time: Timestamp | None = None,
        to_time: Timestamp | None = None,
        filters: dict | None = None,
        include_parents: bool = True,
        limit: int | None = None,
        offset: int | None = None,
        tracking_id: TrackingId | None = None,
        diff_ids: list[str] | None = None,
        include_empty: bool = False,
    ) -> list[EnrichedDiffRoot]:
        final_max_depth = config.SETTINGS.database.max_depth_search_hierarchy
        batch_size_limit = int(config.SETTINGS.database.query_size_limit / 10)
        diff_roots = await self._run_get_diff_query(
            base_branch_name=base_branch_name,
            diff_branch_names=diff_branch_names,
            batch_size_limit=batch_size_limit,
            limit=limit,
            from_time=from_time,
            to_time=to_time,
            filters=EnrichedDiffQueryFilters(**dict(filters or {})),
            include_parents=include_parents,
            max_depth=final_max_depth,
            offset=offset or 0,
            tracking_id=tracking_id,
            diff_ids=diff_ids,
        )
        if not include_empty:
            diff_roots = [dr for dr in diff_roots if len(dr.nodes) > 0]
        return diff_roots

    async def get_pairs(
        self,
        base_branch_name: str,
        diff_branch_name: str,
        from_time: Timestamp,
        to_time: Timestamp,
    ) -> list[EnrichedDiffs]:
        max_depth = config.SETTINGS.database.max_depth_search_hierarchy
        batch_size_limit = int(config.SETTINGS.database.query_size_limit / 10)
        diff_branch_roots = await self._run_get_diff_query(
            base_branch_name=base_branch_name,
            diff_branch_names=[diff_branch_name],
            from_time=from_time,
            to_time=to_time,
            max_depth=max_depth,
            batch_size_limit=batch_size_limit,
        )
        diffs_by_uuid = {dbr.uuid: dbr for dbr in diff_branch_roots}
        base_branch_roots = await self._run_get_diff_query(
            base_branch_name=base_branch_name,
            diff_branch_names=[base_branch_name],
            max_depth=max_depth,
            batch_size_limit=batch_size_limit,
            diff_ids=[d.partner_uuid for d in diffs_by_uuid.values() if d.partner_uuid],
        )
        diffs_by_uuid.update({bbr.uuid: bbr for bbr in base_branch_roots})
        diff_pairs = []
        for dbr in diff_branch_roots:
            if dbr.partner_uuid is None:
                continue
            base_branch_diff = diffs_by_uuid[dbr.partner_uuid]
            diff_pairs.append(
                EnrichedDiffs(
                    base_branch_name=base_branch_name,
                    diff_branch_name=diff_branch_name,
                    base_branch_diff=base_branch_diff,
                    diff_branch_diff=dbr,
                )
            )
        return diff_pairs

    async def hydrate_diff_pair(
        self,
        enriched_diffs_metadata: EnrichedDiffsMetadata,
        node_uuids: Iterable[str] | None = None,
    ) -> EnrichedDiffs:
        filters = None
        if node_uuids:
            filters = {"ids": list(node_uuids) if node_uuids is not None else None}
        hydrated_base_diff = await self.get_one(
            diff_branch_name=enriched_diffs_metadata.base_branch_name,
            diff_id=enriched_diffs_metadata.base_branch_diff.uuid,
            filters=filters,
        )
        hydrated_branch_diff = await self.get_one(
            diff_branch_name=enriched_diffs_metadata.diff_branch_name,
            diff_id=enriched_diffs_metadata.diff_branch_diff.uuid,
            filters=filters,
        )
        return EnrichedDiffs(
            base_branch_name=enriched_diffs_metadata.base_branch_name,
            diff_branch_name=enriched_diffs_metadata.diff_branch_name,
            base_branch_diff=hydrated_base_diff,
            diff_branch_diff=hydrated_branch_diff,
        )

    async def get_one(
        self,
        diff_branch_name: str,
        tracking_id: TrackingId | None = None,
        diff_id: str | None = None,
        filters: dict | None = None,
        include_parents: bool = True,
    ) -> EnrichedDiffRoot:
        enriched_diffs = await self.get(
            base_branch_name=registry.default_branch,
            diff_branch_names=[diff_branch_name],
            tracking_id=tracking_id,
            diff_ids=[diff_id] if diff_id else None,
            filters=filters,
            include_parents=include_parents,
            include_empty=True,
        )
        error_str = f"branch {diff_branch_name}"
        if tracking_id:
            error_str += f" with tracking_id {tracking_id.serialize()}"
        if diff_id:
            error_str += f" with ID {diff_id}"
        if len(enriched_diffs) == 0:
            raise ResourceNotFoundError(f"Cannot find diff for {error_str}")
        if len(enriched_diffs) > 1:
            raise ResourceNotFoundError(f"Multiple diffs for {error_str}")
        return enriched_diffs[0]

    def _get_node_create_request_batch(
        self, enriched_diffs: EnrichedDiffs
    ) -> Generator[list[EnrichedNodeCreateRequest], None, None]:
        node_requests = []
        for diff_root in (enriched_diffs.base_branch_diff, enriched_diffs.diff_branch_diff):
            size_count = 0
            for node in diff_root.nodes:
                node_size_count = node.num_properties
                if size_count + node_size_count < self.max_save_batch_size:
                    node_requests.append(EnrichedNodeCreateRequest(node=node, root_uuid=diff_root.uuid))
                    size_count += node_size_count
                else:
                    log.info(f"Num nodes in batch: {len(node_requests)}, num properties in batch: {size_count}")
                    yield node_requests
                    size_count = node_size_count
                    node_requests = [EnrichedNodeCreateRequest(node=node, root_uuid=diff_root.uuid)]
        if node_requests:
            log.info(f"Num nodes in batch: {len(node_requests)}, num properties in batch: {size_count}")
            yield node_requests

    @retry_db_transaction(name="enriched_diff_metadata_save")
    async def _save_root_metadata(self, enriched_diffs: EnrichedDiffsMetadata) -> None:
        log.info("Updating diff metadata...")
        root_query = await EnrichedDiffRootsUpsertQuery.init(db=self.db, enriched_diffs=enriched_diffs)
        await root_query.execute(db=self.db)
        log.info("Diff metadata updated.")

    async def _save_node_batch(self, node_create_batch: list[EnrichedNodeCreateRequest]) -> None:
        node_query = await EnrichedNodeBatchCreateQuery.init(db=self.db, node_create_batch=node_create_batch)
        try:
            await node_query.execute(db=self.db)
        except TransientError as exc:
            if not exc.code or "OutOfMemoryError".lower() not in str(exc.code).lower():
                raise
            log.exception("Database memory error during save. Trying smaller transactions")
            for node_request in node_create_batch:
                log.info(
                    f"Updating node {node_request.node.uuid}, num_properties={node_request.node.num_properties}..."
                )
                single_node_query = await EnrichedNodeBatchCreateQuery.init(
                    db=self.db, node_create_batch=[node_request]
                )
                await single_node_query.execute(db=self.db)

    @retry_db_transaction(name="enriched_diff_hierarchy_update")
    async def _run_hierarchy_links_update_query(self, diff_root_uuid: str, diff_nodes: list[EnrichedDiffNode]) -> None:
        log.info(f"Updating diff hierarchy links, num_nodes={len(diff_nodes)}")
        link_query = await EnrichedNodesLinkQuery.init(db=self.db, diff_root_uuid=diff_root_uuid, diff_nodes=diff_nodes)
        await link_query.execute(db=self.db)

    async def _update_hierarchy_links(self, enriched_diffs: EnrichedDiffs) -> None:
        for diff_root in (enriched_diffs.base_branch_diff, enriched_diffs.diff_branch_diff):
            nodes_to_update = []
            for node in diff_root.nodes:
                if any(r.nodes for r in node.relationships):
                    nodes_to_update.append(node)
                if len(nodes_to_update) >= config.SETTINGS.database.query_size_limit:
                    await self._run_hierarchy_links_update_query(
                        diff_root_uuid=diff_root.uuid, diff_nodes=nodes_to_update
                    )
                    nodes_to_update = []
            if nodes_to_update:
                await self._run_hierarchy_links_update_query(diff_root_uuid=diff_root.uuid, diff_nodes=nodes_to_update)

    async def _update_summary_counts(self, diff_root: EnrichedDiffRoot) -> None:
        max_nodes_limit = config.SETTINGS.database.query_size_limit
        num_nodes = len(diff_root.nodes)
        if diff_root.exists_on_database and num_nodes < max_nodes_limit:
            await self.add_summary_counts(
                diff_branch_name=diff_root.diff_branch_name,
                diff_id=diff_root.uuid,
                node_uuids=None,
            )
            return
        node_uuids: list[str] = []
        for diff_node in diff_root.nodes:
            node_uuids.append(diff_node.uuid)
            if len(node_uuids) >= max_nodes_limit:
                await self.add_summary_counts(
                    diff_branch_name=diff_root.diff_branch_name,
                    diff_id=diff_root.uuid,
                    node_uuids=node_uuids,
                )
                node_uuids = []
        if node_uuids:
            await self.add_summary_counts(
                diff_branch_name=diff_root.diff_branch_name,
                diff_id=diff_root.uuid,
                node_uuids=node_uuids,
            )

    async def save(self, enriched_diffs: EnrichedDiffs | EnrichedDiffsMetadata, do_summary_counts: bool = True) -> None:
        # metadata-only update
        if not isinstance(enriched_diffs, EnrichedDiffs):
            await self._save_root_metadata(enriched_diffs=enriched_diffs)
            return

        count_nodes_remaining = len(enriched_diffs.base_branch_diff.nodes) + len(enriched_diffs.diff_branch_diff.nodes)
        log.info(f"Saving diff (num_nodes={count_nodes_remaining})...")
        for batch_num, node_create_batch in enumerate(
            self._get_node_create_request_batch(enriched_diffs=enriched_diffs)
        ):
            log.info(f"Saving node batch #{batch_num}...")
            await self._save_node_batch(node_create_batch=node_create_batch)
            count_nodes_remaining -= len(node_create_batch)
            log.info(f"Batch saved. {count_nodes_remaining=}")
        await self._update_hierarchy_links(enriched_diffs=enriched_diffs)
        if do_summary_counts:
            await self._update_summary_counts(diff_root=enriched_diffs.diff_branch_diff)
        await self._save_root_metadata(enriched_diffs=enriched_diffs)

    async def summary(
        self,
        base_branch_name: str,
        diff_branch_names: list[str],
        from_time: Timestamp | None = None,
        to_time: Timestamp | None = None,
        tracking_id: TrackingId | None = None,
        filters: dict | None = None,
    ) -> DiffSummaryCounters | None:
        query = await DiffSummaryQuery.init(
            db=self.db,
            base_branch_name=base_branch_name,
            diff_branch_names=diff_branch_names,
            filters=EnrichedDiffQueryFilters(**dict(filters or {})),
            from_time=from_time,
            to_time=to_time,
            tracking_id=tracking_id,
        )
        await query.execute(db=self.db)
        return query.get_summary()

    async def delete_diff_roots(self, diff_root_uuids: list[str]) -> None:
        query = await EnrichedDiffDeleteQuery.init(db=self.db, enriched_diff_root_uuids=diff_root_uuids)
        await query.execute(db=self.db)

    async def get_time_ranges(
        self,
        base_branch_name: str,
        diff_branch_name: str,
        from_time: Timestamp,
        to_time: Timestamp,
    ) -> list[TimeRange]:
        query = await EnrichedDiffTimeRangeQuery.init(
            db=self.db,
            base_branch_name=base_branch_name,
            diff_branch_name=diff_branch_name,
            from_time=from_time,
            to_time=to_time,
        )
        await query.execute(db=self.db)
        return await query.get_time_ranges()

    async def get_diff_pairs_metadata(
        self,
        diff_branch_names: list[str] | None = None,
        base_branch_names: list[str] | None = None,
        from_time: Timestamp | None = None,
        to_time: Timestamp | None = None,
        tracking_id: TrackingId | None = None,
    ) -> list[EnrichedDiffsMetadata]:
        if diff_branch_names and base_branch_names:
            diff_branch_names += base_branch_names
        empty_roots = await self.get_roots_metadata(
            diff_branch_names=diff_branch_names,
            base_branch_names=base_branch_names,
            from_time=from_time,
            to_time=to_time,
            tracking_id=tracking_id,
        )
        roots_by_id = {root.uuid: root for root in empty_roots}
        pairs: list[EnrichedDiffsMetadata] = []
        for branch_root in empty_roots:
            if branch_root.base_branch_name == branch_root.diff_branch_name or branch_root.partner_uuid is None:
                continue
            base_root = roots_by_id[branch_root.partner_uuid]
            pairs.append(
                EnrichedDiffsMetadata(
                    base_branch_name=branch_root.base_branch_name,
                    diff_branch_name=branch_root.diff_branch_name,
                    base_branch_diff=base_root,
                    diff_branch_diff=branch_root,
                )
            )
        return pairs

    async def get_roots_metadata(
        self,
        diff_branch_names: list[str] | None = None,
        base_branch_names: list[str] | None = None,
        from_time: Timestamp | None = None,
        to_time: Timestamp | None = None,
        tracking_id: TrackingId | None = None,
    ) -> list[EnrichedDiffRootMetadata]:
        query = await EnrichedDiffRootsMetadataQuery.init(
            db=self.db,
            diff_branch_names=diff_branch_names,
            base_branch_names=base_branch_names,
            from_time=from_time,
            to_time=to_time,
            tracking_id=tracking_id,
        )
        await query.execute(db=self.db)
        diff_roots = []
        for neo4j_node in query.get_root_nodes_metadata():
            diff_roots.append(self.deserializer.build_diff_root_metadata(root_node=neo4j_node))
        return diff_roots

    async def diff_has_conflicts(
        self,
        diff_branch_name: str,
        tracking_id: TrackingId | None = None,
        diff_id: str | None = None,
    ) -> bool:
        query = await EnrichedDiffHasConflictQuery.init(
            db=self.db, diff_branch_name=diff_branch_name, tracking_id=tracking_id, diff_id=diff_id
        )
        await query.execute(db=self.db)
        return await query.has_conflict()

    async def get_conflict_by_id(self, conflict_id: str) -> EnrichedDiffConflict:
        query = await EnrichedDiffConflictQuery.init(db=self.db, conflict_id=conflict_id)
        await query.execute(db=self.db)
        conflict_node = await query.get_conflict_node()
        if not conflict_node:
            raise ResourceNotFoundError(f"No conflict with id {conflict_id}")
        return self.deserializer.deserialize_conflict(diff_conflict_node=conflict_node)

    async def update_conflict_by_id(
        self, conflict_id: str, selection: ConflictSelection | None
    ) -> EnrichedDiffConflict:
        query = await EnrichedDiffConflictUpdateQuery.init(db=self.db, conflict_id=conflict_id, selection=selection)
        await query.execute(db=self.db)
        conflict_node = await query.get_conflict_node()
        if not conflict_node:
            raise ResourceNotFoundError(f"No conflict with id {conflict_id}")
        return self.deserializer.deserialize_conflict(diff_conflict_node=conflict_node)

    async def get_all_conflicts_for_diff(
        self,
        diff_branch_name: str,
        tracking_id: TrackingId | None = None,
        diff_id: str | None = None,
    ) -> AsyncGenerator[tuple[str, EnrichedDiffConflict], None]:
        query = await EnrichedDiffAllConflictsQuery.init(
            db=self.db, diff_branch_name=diff_branch_name, tracking_id=tracking_id, diff_id=diff_id
        )
        await query.execute(db=self.db)
        for conflict_path, conflict_node in query.get_conflict_paths_and_nodes():
            yield (conflict_path, self.deserializer.deserialize_conflict(diff_conflict_node=conflict_node))

    async def get_node_field_summaries(
        self, diff_branch_name: str, tracking_id: TrackingId | None = None, diff_id: str | None = None
    ) -> list[NodeDiffFieldSummary]:
        query = await EnrichedDiffNodeFieldSummaryQuery.init(
            db=self.db, diff_branch_name=diff_branch_name, tracking_id=tracking_id, diff_id=diff_id
        )
        await query.execute(db=self.db)
        return await query.get_field_summaries()

    async def mark_tracking_ids_merged(self, tracking_ids: list[TrackingId]) -> None:
        query = await EnrichedDiffMergedTrackingIdQuery.init(db=self.db, tracking_ids=tracking_ids)
        await query.execute(db=self.db)

    async def get_num_changes_in_time_range_by_branch(
        self, branch_names: list[str], from_time: Timestamp, to_time: Timestamp
    ) -> dict[str, int]:
        query = await DiffCountChanges.init(db=self.db, branch_names=branch_names, diff_from=from_time, diff_to=to_time)
        await query.execute(db=self.db)
        return query.get_num_changes_by_branch()

    async def get_node_field_specifiers(self, diff_id: str) -> dict[str, set[str]]:
        limit = config.SETTINGS.database.query_size_limit
        offset = 0
        specifiers: dict[str, set[str]] = defaultdict(set)
        while True:
            query = await EnrichedDiffFieldSpecifiersQuery.init(db=self.db, diff_id=diff_id, offset=offset, limit=limit)
            await query.execute(db=self.db)
            has_data = False
            for field_specifier_tuple in query.get_node_field_specifier_tuples():
                specifiers[field_specifier_tuple[0]].add(field_specifier_tuple[1])
                has_data = True
            if not has_data:
                break
            offset += limit
        return specifiers

    async def add_summary_counts(
        self,
        diff_branch_name: str,
        tracking_id: TrackingId | None = None,
        diff_id: str | None = None,
        node_uuids: list[str] | None = None,
    ) -> None:
        await self._add_field_summary_counts(
            diff_branch_name=diff_branch_name,
            tracking_id=tracking_id,
            diff_id=diff_id,
            node_uuids=node_uuids,
        )
        await self._add_node_summary_counts(
            diff_branch_name=diff_branch_name,
            tracking_id=tracking_id,
            diff_id=diff_id,
            node_uuids=node_uuids,
        )

    @retry_db_transaction(name="enriched_diff_field_summary_counts")
    async def _add_field_summary_counts(
        self,
        diff_branch_name: str,
        tracking_id: TrackingId | None = None,
        diff_id: str | None = None,
        node_uuids: list[str] | None = None,
    ) -> None:
        log.info("Updating field summary counts...")
        query = await DiffFieldsSummaryCountsEnricherQuery.init(
            db=self.db,
            diff_branch_name=diff_branch_name,
            tracking_id=tracking_id,
            diff_id=diff_id,
            node_uuids=node_uuids,
        )
        await query.execute(db=self.db)
        log.info("Field summary counts updated.")

    @retry_db_transaction(name="enriched_diff_node_summary_counts")
    async def _add_node_summary_counts(
        self,
        diff_branch_name: str,
        tracking_id: TrackingId | None = None,
        diff_id: str | None = None,
        node_uuids: list[str] | None = None,
    ) -> None:
        log.info("Updating node summary counts...")
        query = await DiffNodesSummaryCountsEnricherQuery.init(
            db=self.db,
            diff_branch_name=diff_branch_name,
            tracking_id=tracking_id,
            diff_id=diff_id,
            node_uuids=node_uuids,
        )
        await query.execute(db=self.db)
        log.info("node summary counts updated.")



# Generated by "invoke backend.generate", do not edit directly

from __future__ import annotations

from pydantic import Field

from infrahub.core.schema.basenode_schema import BaseNodeSchema


class GeneratedGenericSchema(BaseNodeSchema):
    hierarchical: bool = Field(
        default=False,
        description="Defines if the Generic support the hierarchical mode.",
        json_schema_extra={"update": "validate_constraint"},
    )
    generate_profile: bool = Field(
        default=True,
        description="Indicate if a profile schema should be generated for this schema",
        json_schema_extra={"update": "validate_constraint"},
    )
    generate_template: bool = Field(
        default=False,
        description="Indicate if an object template schema should be generated for this schema",
        json_schema_extra={"update": "allowed"},
    )
    used_by: list[str] = Field(
        default_factory=list,
        description="List of Nodes that are referencing this Generic",
        json_schema_extra={"update": "not_applicable"},
    )



from __future__ import annotations

import hashlib
import keyword
import os
from dataclasses import asdict, dataclass
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, Iterable, Literal, overload

from infrahub_sdk.utils import compare_lists, intersection
from pydantic import field_validator

from infrahub.core.constants import RelationshipCardinality, RelationshipKind
from infrahub.core.models import HashableModel, HashableModelDiff

from .attribute_schema import AttributeSchema
from .generated.base_node_schema import GeneratedBaseNodeSchema
from .relationship_schema import RelationshipSchema

if TYPE_CHECKING:
    from typing_extensions import Self

    from infrahub.core.schema import GenericSchema, NodeSchema
    from infrahub.core.schema.schema_branch import SchemaBranch


NODE_METADATA_ATTRIBUTES = ["_source", "_owner"]
INHERITED = "INHERITED"

OPTIONAL_TEXT_FIELDS = [
    "default_filter",
    "description",
    "label",
    "menu_placement",
    "documentation",
    "parent",
    "children",
]


class BaseNodeSchema(GeneratedBaseNodeSchema):
    _exclude_from_hash: list[str] = ["attributes", "relationships"]
    _sort_by: list[str] = ["namespace", "name"]

    @property
    def is_node_schema(self) -> bool:
        return False

    @property
    def is_generic_schema(self) -> bool:
        return False

    @property
    def is_profile_schema(self) -> bool:
        return False

    @property
    def kind(self) -> str:
        if self.namespace == "Attribute":
            return self.name
        return self.namespace + self.name

    @property
    def menu_title(self) -> str:
        return self.label or self.name

    def get_id(self) -> str:
        if self.id:
            return self.id
        raise ValueError(f"id is not defined on {self.kind}")

    def __hash__(self) -> int:
        """Return a hash of the object.
        Be careful hash generated from hash() have a salt by default and they will not be the same across run"""
        return hash(self.get_hash())

    def to_dict(self) -> dict:
        data = self.model_dump(
            exclude_unset=True, exclude_none=True, exclude_defaults=True, exclude={"attributes", "relationships"}
        )
        for field_name, value in data.items():
            if isinstance(value, Enum):
                data[field_name] = value.value
        data["attributes"] = [attr.to_dict() for attr in self.attributes]
        data["relationships"] = [rel.to_dict() for rel in self.relationships]
        return data

    def get_hash(self, display_values: bool = False) -> str:
        """Extend the Hash Calculation to account for attributes and relationships."""

        md5hash = hashlib.md5(usedforsecurity=False)
        md5hash.update(super().get_hash(display_values=display_values).encode())

        for attr_name in sorted(self.attribute_names):
            md5hash.update(self.get_attribute(name=attr_name).get_hash(display_values=display_values).encode())

        for rel_name in sorted(self.relationship_names):
            md5hash.update(self.get_relationship(name=rel_name).get_hash(display_values=display_values).encode())

        return md5hash.hexdigest()

    def diff(self, other: Self) -> HashableModelDiff:
        """Extend the Diff Calculation to account for attributes and relationships."""

        node_diff = super().diff(other=other)

        # Attribute
        attrs_diff = self._diff_element(
            other=other,
            get_func=BaseNodeSchema.get_attribute,
            get_map_func=BaseNodeSchema.get_attributes_name_id_map,
            obj_type=AttributeSchema,
        )
        # Relationships
        rels_diff = self._diff_element(
            other=other,
            get_func=BaseNodeSchema.get_relationship,
            get_map_func=BaseNodeSchema.get_relationship_name_id_map,
            obj_type=RelationshipSchema,
        )

        if attrs_diff.has_diff:
            node_diff.changed["attributes"] = attrs_diff
        if rels_diff.has_diff:
            node_diff.changed["relationships"] = rels_diff

        return node_diff

    def _diff_element(
        self,
        other: Self,
        get_func: Callable,
        get_map_func: Callable,
        obj_type: type[AttributeSchema | RelationshipSchema],
    ) -> HashableModelDiff:
        """The goal of this function is to reduce the amount of code duplicated between Attribute and Relationship to calculate a diff
        The logic is the same for both, except that the functions we are using to access these objects are differents

        To map elements from the local and other objects together, we are using a combinasion of ID and name
        If the same id is present on both we'll use the ID to match the elements on both side
        If the id is not present on either side, we'll try to match with the name

        """
        # Build a mapping between name and id for all element as well as the reverse mapping to make it easy to access the data
        local_map: dict[str, str] = get_map_func(self)
        other_map: dict[str, str] = get_map_func(other)

        reversed_map_local = dict(map(reversed, local_map.items()))
        reversed_map_other = dict(map(reversed, other_map.items()))

        # Identify which elements are using the same id on both sides
        clean_local_ids = [id for id in local_map.values() if id is not None and id != INHERITED]
        clean_other_ids = [id for id in other_map.values() if id is not None and id != INHERITED]
        shared_ids = intersection(list1=clean_local_ids, list2=clean_other_ids)

        # Identify which elements are present on both side based on the name
        local_names = [name for name, id in local_map.items() if id not in shared_ids]
        other_names = [name for name, id in other_map.items() if id not in shared_ids]
        present_both, present_local, present_other = compare_lists(list1=local_names, list2=other_names)

        elements_diff = HashableModelDiff()
        if present_local:
            elements_diff.added = dict.fromkeys(present_local)
        if present_other:
            elements_diff.removed = dict.fromkeys(present_other)

        # Process element b
        for name in sorted(present_both):
            # If the element doesn't have an ID on either side
            # this most likely means it was added recently from the internal schema.
            if os.environ.get("PYTEST_RUNNING", "") != "true" and local_map[name] is None and other_map[name] is None:
                elements_diff.added[name] = None
                continue
            local_element: obj_type = get_func(self, name=name)
            other_element: obj_type = get_func(other, name=name)
            element_diff = local_element.diff(other_element)
            if element_diff.has_diff:
                elements_diff.changed[name] = element_diff

        for element_id in shared_ids:
            local_element: obj_type = get_func(self, name=reversed_map_local[element_id])
            other_element: obj_type = get_func(other, name=reversed_map_other[element_id])
            element_diff = local_element.diff(other_element)
            if element_diff.has_diff:
                elements_diff.changed[reversed_map_local[element_id]] = element_diff

        return elements_diff

    @overload
    def get_field(self, name: str, raise_on_error: Literal[True] = True) -> AttributeSchema | RelationshipSchema: ...

    @overload
    def get_field(
        self, name: str, raise_on_error: Literal[False] = False
    ) -> AttributeSchema | RelationshipSchema | None: ...

    def get_field(self, name: str, raise_on_error: bool = True) -> AttributeSchema | RelationshipSchema | None:
        if field := self.get_attribute_or_none(name=name):
            return field

        if field := self.get_relationship_or_none(name=name):
            return field

        if raise_on_error:
            raise ValueError(f"Unable to find the field {name}")

        return None

    def get_attribute(self, name: str) -> AttributeSchema:
        for item in self.attributes:
            if item.name == name:
                return item

        raise ValueError(f"Unable to find the attribute {name}")

    def get_attribute_or_none(self, name: str) -> AttributeSchema | None:
        for item in self.attributes:
            if item.name == name:
                return item
        return None

    def get_attribute_by_id(self, id: str) -> AttributeSchema:
        for item in self.attributes:
            if item.id == id:
                return item

        raise ValueError(f"Unable to find the attribute with the ID: {id}")

    def get_relationship(self, name: str) -> RelationshipSchema:
        for item in self.relationships:
            if item.name == name:
                return item
        raise ValueError(f"Unable to find the relationship {name}")

    def get_relationship_by_id(self, id: str) -> RelationshipSchema:
        for item in self.relationships:
            if item.id == id:
                return item

        raise ValueError(f"Unable to find the relationship with the ID: {id}")

    def get_relationship_or_none(self, name: str) -> RelationshipSchema | None:
        for item in self.relationships:
            if item.name == name:
                return item
        return None

    @overload
    def get_relationship_by_identifier(self, id: str, raise_on_error: Literal[True] = True) -> RelationshipSchema: ...

    @overload
    def get_relationship_by_identifier(
        self, id: str, raise_on_error: Literal[False] = False
    ) -> RelationshipSchema | None: ...

    def get_relationship_by_identifier(self, id: str, raise_on_error: bool = True) -> RelationshipSchema | None:
        for item in self.relationships:
            if item.identifier == id:
                return item

        if not raise_on_error:
            return None

        raise ValueError(f"Unable to find the relationship {id}")

    def get_relationships_by_identifier(self, id: str) -> list[RelationshipSchema]:
        """Return a list of relationship instead of a single one"""
        rels: list[RelationshipSchema] = []
        for item in self.relationships:
            if item.identifier == id:
                rels.append(item)

        return rels

    def get_relationships_of_kind(self, relationship_kinds: Iterable[RelationshipKind]) -> list[RelationshipSchema]:
        return [r for r in self.relationships if r.kind in relationship_kinds]

    def get_attributes_name_id_map(self) -> dict[str, str]:
        name_id_map = {}
        for attr in self.attributes:
            name_id_map[attr.name] = INHERITED if attr.inherited else attr.id
        return name_id_map

    def get_relationship_name_id_map(self) -> dict[str, str]:
        name_id_map = {}
        for rel in self.relationships:
            name_id_map[rel.name] = INHERITED if rel.inherited else rel.id
        return name_id_map

    @property
    def has_parent_relationship(self) -> bool:
        for rel in self.relationships:
            if rel.kind == RelationshipKind.PARENT:
                return True
        return False

    @property
    def valid_input_names(self) -> list[str]:
        return self.attribute_names + self.relationship_names + NODE_METADATA_ATTRIBUTES

    @property
    def valid_local_names(self) -> list[str]:
        return self.local_attribute_names + self.local_relationship_names + NODE_METADATA_ATTRIBUTES

    @property
    def attribute_names(self) -> list[str]:
        return [item.name for item in self.attributes]

    @property
    def relationship_names(self) -> list[str]:
        return [item.name for item in self.relationships]

    @property
    def mandatory_input_names(self) -> list[str]:
        return self.mandatory_attribute_names + self.mandatory_relationship_names

    @property
    def mandatory_attribute_names(self) -> list[str]:
        return [item.name for item in self.attributes if not item.optional and item.default_value is None]

    @property
    def mandatory_relationship_names(self) -> list[str]:
        return [item.name for item in self.relationships if not item.optional]

    @property
    def local_attributes(self) -> list[AttributeSchema]:
        return [item for item in self.attributes if not item.inherited]

    @property
    def local_attribute_names(self) -> list[str]:
        return [item.name for item in self.local_attributes]

    @property
    def local_relationships(self) -> list[RelationshipSchema]:
        return [item for item in self.relationships if not item.inherited]

    @property
    def local_relationship_names(self) -> list[str]:
        return [item.name for item in self.local_relationships]

    @property
    def unique_attributes(self) -> list[AttributeSchema]:
        return [item for item in self.attributes if item.unique]

    @classmethod
    def convert_path_to_graphql_fields(cls, path: str) -> dict:
        subpaths = path.split("__", maxsplit=1)
        fields = {}
        if len(subpaths) == 1:
            fields[subpaths[0]] = None
        elif len(subpaths) == 2:
            fields[subpaths[0]] = cls.convert_path_to_graphql_fields(path=subpaths[1])
        return fields

    def generate_fields_for_display_label(self) -> dict | None:
        """Generate a dictionary containing the list of fields that are required
        to generate the display_label.

        If display_labels is not defined, we return None which equal to everything.
        """

        if not self.display_labels:
            return None

        fields: dict[str, str | None | dict[str, None]] = {}
        for item in self.display_labels:
            fields.update(self.convert_path_to_graphql_fields(path=item))
        return fields

    def generate_fields_for_hfid(self) -> dict | None:
        """Generate a dictionary containing the list of fields that are required
        to generate the hfid.

        If display_labels is not defined, we return None which equal to everything.
        """

        if not self.human_friendly_id:
            return None

        fields: dict[str, str | None | dict[str, None]] = {}
        for item in self.human_friendly_id:
            fields.update(self.convert_path_to_graphql_fields(path=item))
        return fields

    @field_validator("name")
    @classmethod
    def name_is_not_keyword(cls, value: str) -> str:
        if keyword.iskeyword(value):
            raise ValueError(f"Name can not be set to a reserved keyword '{value}' is not allowed.")

        return value

    def parse_schema_path(self, path: str, schema: SchemaBranch | None = None) -> SchemaAttributePath:
        schema_path = SchemaAttributePath()
        relationship_piece: str | None = None
        attribute_piece: str | None = None
        property_piece: str | None = None

        path_parts = path.split("__")
        if path_parts[0] in self.relationship_names:
            relationship_piece = path_parts[0]
            attribute_piece = path_parts[1] if len(path_parts) > 1 else None
            property_piece = path_parts[2] if len(path_parts) > 2 else None
        elif path_parts[0] in self.attribute_names:
            attribute_piece = path_parts[0]
            property_piece = path_parts[1] if len(path_parts) > 1 else None
        elif path_parts[0] == "parent" and schema:
            relationship_piece = path_parts[0]
            peer_schema_name = getattr(self, path_parts[0])
            schema_path.relationship_schema = RelationshipSchema(
                name="parent", peer=peer_schema_name, cardinality=RelationshipCardinality.ONE, optional=True
            )
            schema_path.related_schema = schema.get(name=peer_schema_name, duplicate=True)
            attribute_piece = path_parts[1] if len(path_parts) > 1 else None
            property_piece = path_parts[2] if len(path_parts) > 2 else None
        else:
            raise AttributePathParsingError(f"{path} is invalid on schema {self.kind}")

        if relationship_piece and not schema:
            raise AttributePathParsingError("schema must be provided in order to check a path with a relationship")

        if relationship_piece and not schema_path.related_schema:
            relationship_schema = self.get_relationship(name=path_parts[0])
            schema_path.relationship_schema = relationship_schema
            schema_path.related_schema = schema.get(name=relationship_schema.peer, duplicate=True)

        if attribute_piece:
            schema_to_check = schema_path.related_schema or self
            if attribute_piece not in schema_to_check.attribute_names:
                raise AttributePathParsingError(f"{attribute_piece} is not a valid attribute of {schema_to_check.kind}")
            schema_path.attribute_schema = schema_to_check.get_attribute(name=attribute_piece)

            if property_piece:
                attr_class = schema_path.attribute_schema.get_class()
                if property_piece not in attr_class.get_allowed_property_in_path():
                    raise AttributePathParsingError(
                        f"{property_piece} is not a valid property of {schema_path.attribute_schema.name}"
                    )
                schema_path.attribute_property_name = property_piece

        return schema_path

    def get_unique_constraint_schema_attribute_paths(
        self,
        schema_branch: SchemaBranch,
    ) -> list[SchemaUniquenessConstraintPath]:
        if self.uniqueness_constraints is None:
            return []

        uniqueness_constraint_paths = []

        for uniqueness_path_group in self.uniqueness_constraints:
            attributes_paths = [
                self.parse_schema_path(path=uniqueness_path_part, schema=schema_branch)
                for uniqueness_path_part in uniqueness_path_group
            ]
            uniqueness_constraint_type = self.get_uniqueness_constraint_type(
                uniqueness_constraint=set(uniqueness_path_group), schema_branch=schema_branch
            )
            uniqueness_constraint_path = SchemaUniquenessConstraintPath(
                attributes_paths=attributes_paths, typ=uniqueness_constraint_type
            )
            uniqueness_constraint_paths.append(uniqueness_constraint_path)

        return uniqueness_constraint_paths

    def convert_hfid_to_uniqueness_constraint(self, schema_branch: SchemaBranch) -> list[str] | None:
        if self.human_friendly_id is None:
            return None

        uniqueness_constraint = []
        for item in self.human_friendly_id:
            schema_attribute_path = self.parse_schema_path(path=item, schema=schema_branch)
            if schema_attribute_path.is_type_attribute:
                uniqueness_constraint.append(item)
            elif schema_attribute_path.is_type_relationship:
                uniqueness_constraint.append(schema_attribute_path.relationship_schema.name)
        return uniqueness_constraint

    def get_uniqueness_constraint_type(
        self, uniqueness_constraint: set[str], schema_branch: SchemaBranch
    ) -> UniquenessConstraintType:
        hfid = self.convert_hfid_to_uniqueness_constraint(schema_branch=schema_branch)
        if hfid is None:
            return UniquenessConstraintType.STANDARD
        hfid_set = set(hfid)
        if uniqueness_constraint == hfid_set:
            return UniquenessConstraintType.HFID
        if uniqueness_constraint <= hfid_set:
            return UniquenessConstraintType.SUBSET_OF_HFID
        return UniquenessConstraintType.STANDARD

    def update(self, other: HashableModel) -> Self:
        super().update(other=other)

        # Allow to specify empty string to remove existing fields values
        for field_name in OPTIONAL_TEXT_FIELDS:
            if getattr(other, field_name, None) == "":  # noqa: PLC1901
                setattr(self, field_name, None)

        return self


@dataclass
class SchemaUniquenessConstraintPath:
    attributes_paths: list[SchemaAttributePath]
    typ: UniquenessConstraintType


class UniquenessConstraintType(Enum):
    HFID = "HFID"
    SUBSET_OF_HFID = "SUBSET_OF_HFID"
    STANDARD = "STANDARD"


@dataclass
class UniquenessConstraintViolation:
    nodes_ids: set[str]
    fields: list[str]
    typ: UniquenessConstraintType


@dataclass
class SchemaAttributePath:
    relationship_schema: RelationshipSchema | None = None
    related_schema: NodeSchema | GenericSchema | None = None
    attribute_schema: AttributeSchema | None = None
    attribute_property_name: str | None = None

    @property
    def is_type_attribute(self) -> bool:
        return bool(self.attribute_schema and not self.related_schema and not self.relationship_schema)

    @property
    def is_type_relationship(self) -> bool:
        return bool(self.relationship_schema and self.related_schema)

    @property
    def has_property(self) -> bool:
        return bool(self.attribute_property_name)

    @property
    def active_relationship_schema(self) -> RelationshipSchema:
        if self.relationship_schema:
            return self.relationship_schema
        raise AttributePathParsingError("A relation_schema was expected but not found")

    @property
    def active_attribute_schema(self) -> AttributeSchema:
        if self.attribute_schema:
            return self.attribute_schema
        raise AttributePathParsingError("An attribute_schema was expected but not found")

    @property
    def active_attribute_property_name(self) -> str:
        if self.attribute_property_name:
            return self.attribute_property_name
        raise AttributePathParsingError("An attribute_property_name was expected but not found")

    @property
    def attribute_path_as_str(self) -> str:
        return self.active_attribute_schema.name + "__" + self.active_attribute_property_name


@dataclass
class SchemaAttributePathValue(SchemaAttributePath):
    value: Any = None

    @classmethod
    def from_schema_attribute_path(
        cls, schema_attribute_path: SchemaAttributePath, value: Any
    ) -> SchemaAttributePathValue:
        return cls(**asdict(schema_attribute_path), value=value)


class AttributePathParsingError(Exception): ...



from infrahub.core.constants import PathType, RelationshipCardinality
from infrahub.core.constants.database import DatabaseEdgeType
from infrahub.core.diff.model.diff import BranchChanges, DataConflict, ModifiedPathType
from infrahub.database import InfrahubDatabase

from .model.path import (
    EnrichedDiffAttribute,
    EnrichedDiffNode,
    EnrichedDiffProperty,
    EnrichedDiffRelationship,
    EnrichedDiffRoot,
    EnrichedDiffSingleRelationship,
)


class DiffConflictsExtractor:
    def __init__(self, db: InfrahubDatabase) -> None:
        self.db = db
        self._base_branch_name: str | None = None
        self._diff_branch_name: str | None = None
        self._conflict_ids: set[str] | None = None

    @property
    def base_branch_name(self) -> str:
        if not self._base_branch_name:
            raise RuntimeError("could not identify base branch")
        return self._base_branch_name

    @property
    def diff_branch_name(self) -> str:
        if not self._diff_branch_name:
            raise RuntimeError("could not identify diff branch")
        return self._diff_branch_name

    @property
    def conflict_ids(self) -> set[str] | None:
        return self._conflict_ids

    async def get_data_conflicts(
        self, enriched_diff_root: EnrichedDiffRoot, conflict_ids: set[str] | None = None
    ) -> list[DataConflict]:
        self._base_branch_name = enriched_diff_root.base_branch_name
        self._diff_branch_name = enriched_diff_root.diff_branch_name
        self._conflict_ids = conflict_ids
        conflicts: list[DataConflict] = []
        for node in enriched_diff_root.nodes:
            conflicts.extend(self._build_node_conflicts(diff_node=node))
        return conflicts

    def _build_node_conflicts(self, diff_node: EnrichedDiffNode) -> list[DataConflict]:
        conflicts: list[DataConflict] = []
        node_conflict = self._get_node_conflict(diff_node=diff_node)
        if node_conflict:
            conflicts.append(node_conflict)
        for diff_attribute in diff_node.attributes:
            conflicts.extend(self._build_attribute_conflicts(diff_node=diff_node, diff_attribute=diff_attribute))
        for diff_relationship in diff_node.relationships:
            conflicts.extend(
                self._build_relationship_conflicts(
                    diff_node=diff_node,
                    diff_relationship=diff_relationship,
                )
            )
        return conflicts

    def _get_node_conflict(self, diff_node: EnrichedDiffNode) -> DataConflict | None:
        if not diff_node.conflict:
            return None
        if self.conflict_ids and diff_node.conflict.uuid not in self.conflict_ids:
            return None
        path = f"{ModifiedPathType.DATA.value}/{diff_node.uuid}"
        branch_changes = [
            BranchChanges(
                branch=self.base_branch_name,
                action=diff_node.conflict.base_branch_action,
            ),
            BranchChanges(
                branch=self.diff_branch_name,
                action=diff_node.conflict.diff_branch_action,
            ),
        ]
        return DataConflict(
            name="",
            type=ModifiedPathType.DATA.value,
            kind=diff_node.kind,
            id=diff_node.uuid,
            change_type=PathType.NODE.value,
            path=path,
            conflict_path=path,
            path_type=PathType.NODE,
            property_name=None,
            changes=branch_changes,
            conflict_id=diff_node.conflict.uuid,
        )

    def _build_attribute_conflicts(
        self,
        diff_node: EnrichedDiffNode,
        diff_attribute: EnrichedDiffAttribute,
    ) -> list[DataConflict]:
        conflicts: list[DataConflict] = []
        for diff_property in diff_attribute.properties:
            property_conflict = self._get_property_conflict(
                diff_node=diff_node,
                diff_property=diff_property,
                parent_name=diff_attribute.name,
                path_type=PathType.ATTRIBUTE,
            )
            if property_conflict:
                conflicts.append(property_conflict)
        return conflicts

    def _build_relationship_conflicts(
        self,
        diff_node: EnrichedDiffNode,
        diff_relationship: EnrichedDiffRelationship,
    ) -> list[DataConflict]:
        conflicts: list[DataConflict] = []
        if not diff_relationship.relationships:
            return conflicts
        node_schema = self.db.schema.get(name=diff_node.kind, branch=self.diff_branch_name, duplicate=False)
        relationship_schema = node_schema.get_relationship(name=diff_relationship.name)
        is_cardinality_one = relationship_schema.cardinality is RelationshipCardinality.ONE
        for diff_element in diff_relationship.relationships:
            conflicts.extend(
                self._add_relationship_conflicts_for_one_peer(
                    diff_node=diff_node,
                    diff_relationship=diff_relationship,
                    diff_element=diff_element,
                    is_cardinality_one=is_cardinality_one,
                )
            )
        return conflicts

    def _add_relationship_conflicts_for_one_peer(
        self,
        diff_node: EnrichedDiffNode,
        diff_relationship: EnrichedDiffRelationship,
        diff_element: EnrichedDiffSingleRelationship,
        is_cardinality_one: bool,
    ) -> list[DataConflict]:
        conflicts: list[DataConflict] = []
        if is_cardinality_one:
            path_type = PathType.RELATIONSHIP_ONE
        else:
            path_type = PathType.RELATIONSHIP_MANY
        peer_conflict = self._get_peer_conflict(
            diff_node=diff_node,
            diff_element=diff_element,
            relationship_name=diff_relationship.name,
            path_type=path_type,
        )
        if peer_conflict:
            conflicts.append(peer_conflict)
        for diff_property in diff_element.properties:
            property_conflict = self._get_property_conflict(
                diff_node=diff_node,
                diff_property=diff_property,
                parent_name=diff_relationship.name,
                path_type=path_type,
            )
            if property_conflict:
                conflicts.append(property_conflict)
        return conflicts

    def _get_peer_conflict(
        self,
        diff_element: EnrichedDiffSingleRelationship,
        diff_node: EnrichedDiffNode,
        relationship_name: str,
        path_type: PathType,
    ) -> DataConflict | None:
        if not diff_element.conflict:
            return None
        if self.conflict_ids and diff_element.conflict not in self.conflict_ids:
            return None
        try:
            peer_property = diff_element.get_property(property_type=DatabaseEdgeType.IS_RELATED)
            previous_peer_id = peer_property.previous_value
        except ValueError:
            previous_peer_id = None
        branch_changes = [
            BranchChanges(
                branch=self.base_branch_name,
                action=diff_element.conflict.base_branch_action,
                previous=previous_peer_id,
                new=diff_element.conflict.base_branch_value,
            ),
            BranchChanges(
                branch=self.diff_branch_name,
                action=diff_element.conflict.diff_branch_action,
                previous=previous_peer_id,
                new=diff_element.conflict.diff_branch_value,
            ),
        ]
        return DataConflict(
            name=relationship_name,
            type=ModifiedPathType.DATA.value,
            id=diff_node.uuid,
            kind=diff_node.kind,
            change_type=f"{path_type.value}_value",
            path=diff_element.path_identifier,
            conflict_path=diff_element.path_identifier,
            path_type=path_type,
            property_name=None,
            changes=branch_changes,
            conflict_id=diff_element.conflict.uuid,
        )

    def _get_property_conflict(
        self,
        diff_node: EnrichedDiffNode,
        diff_property: EnrichedDiffProperty,
        parent_name: str,
        path_type: PathType,
    ) -> DataConflict | None:
        if not diff_property.conflict:
            return None
        change_type = PathType.ATTRIBUTE.value
        if diff_property.property_type in (DatabaseEdgeType.HAS_VALUE, DatabaseEdgeType.IS_RELATED):
            change_type = f"{path_type.value}_value"
        else:
            change_type = f"{path_type.value}_property"
        branch_changes = [
            BranchChanges(
                branch=self.base_branch_name,
                action=diff_property.conflict.base_branch_action,
                previous=diff_property.previous_value,
                new=diff_property.conflict.base_branch_value,
            ),
            BranchChanges(
                branch=self.diff_branch_name,
                action=diff_property.conflict.diff_branch_action,
                previous=diff_property.previous_value,
                new=diff_property.conflict.diff_branch_value,
            ),
        ]
        return DataConflict(
            name=parent_name,
            type=ModifiedPathType.DATA.value,
            id=diff_node.uuid,
            kind=diff_node.kind,
            change_type=change_type,
            path=diff_property.path_identifier,
            conflict_path=diff_property.path_identifier,
            path_type=path_type,
            property_name=None
            if diff_property.property_type is DatabaseEdgeType.IS_RELATED
            else diff_property.property_type.value,
            changes=branch_changes,
            conflict_id=diff_property.conflict.uuid,
        )



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, ForwardRef, Optional, Union, get_origin

from pydantic import BaseModel
from typing_extensions import Self

from infrahub.core.query import QueryType

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class GraphPropertyType(str, Enum):
    BOOLEAN = "BOOLEAN"
    STRING = "STRING"
    INTEGER = "INTEGER"
    FLOAT = "FLOAT"
    ZONED_DATETIME = "ZONED DATETIME"


class ConstraintInfo(BaseModel):
    item_name: str
    item_label: str
    property: str


class ConstraintItem(ConstraintInfo):
    type: GraphPropertyType
    mandatory: bool = True

    def get_add_queries(self) -> list[str]:
        raise NotImplementedError()

    def get_drop_queries(self) -> list[str]:
        raise NotImplementedError()


class ConstraintItemNeo4j(ConstraintItem):
    def names(self) -> list[str]:
        if self.mandatory:
            return [self.name_main, self.name_exist]
        return [self.name_main]

    @property
    def name_main(self) -> str:
        raise NotImplementedError()

    @property
    def name_exist(self) -> str:
        raise NotImplementedError()

    def get_query_main_add(self) -> str:
        raise NotImplementedError()

    def get_query_exist_add(self) -> str:
        raise NotImplementedError()

    def get_query_main_drop(self) -> str:
        return f"DROP CONSTRAINT {self.name_main} IF EXISTS"

    def get_query_exist_drop(self) -> str:
        return f"DROP CONSTRAINT {self.name_exist} IF EXISTS"

    def get_add_queries(self) -> list[str]:
        queries = [self.get_query_main_add()]
        if self.mandatory:
            queries.append(self.get_query_exist_add())
        return queries

    def get_drop_queries(self) -> list[str]:
        queries = [self.get_query_main_drop()]
        if self.mandatory:
            queries.append(self.get_query_exist_drop())
        return queries


class ConstraintNodeNeo4j(ConstraintItemNeo4j):
    @property
    def name_main(self) -> str:
        return f"node_{self.item_name}_{self.property}_type"

    @property
    def name_exist(self) -> str:
        if not self.mandatory:
            raise ValueError()
        return f"node_{self.item_name}_{self.property}_exist"

    def get_query_main_add(self) -> str:
        return (
            f"CREATE CONSTRAINT {self.name_main} IF NOT EXISTS "
            f"FOR (n:{self.item_label}) REQUIRE n.{self.property} IS :: {self.type.value}"
        )

    def get_query_exist_add(self) -> str:
        return (
            f"CREATE CONSTRAINT {self.name_exist} IF NOT EXISTS "
            f"FOR (n:{self.item_label}) REQUIRE n.{self.property} IS NOT NULL"
        )


class ConstraintRelNeo4j(ConstraintItemNeo4j):
    @property
    def name_main(self) -> str:
        return f"rel_{self.item_name}_{self.property}_type"

    @property
    def name_exist(self) -> str:
        if not self.mandatory:
            raise ValueError()
        return f"rel_{self.item_name}_{self.property}_exist"

    def get_query_main_add(self) -> str:
        return (
            f"CREATE CONSTRAINT {self.name_main} IF NOT EXISTS "
            f"FOR ()-[r:{self.item_label}]-() REQUIRE r.{self.property} IS :: {self.type.value}"
        )

    def get_query_exist_add(self) -> str:
        return (
            f"CREATE CONSTRAINT {self.name_exist} IF NOT EXISTS "
            f"FOR ()-[r:{self.item_label}]-() REQUIRE r.{self.property} IS NOT NULL"
        )


# ------------------------------------------------------
# Memgraph
# ------------------------------------------------------


class ConstraintItemMemgraph(ConstraintItem):
    def get_query_exist_add(self) -> str:
        raise NotImplementedError()

    def get_query_exist_drop(self) -> str:
        raise NotImplementedError()

    def get_add_queries(self) -> list[str]:
        return [self.get_query_exist_add()]

    def get_drop_queries(self) -> list[str]:
        return [self.get_query_exist_drop()]


class ConstraintNodeMemgraph(ConstraintItemMemgraph):
    def get_query_exist_add(self) -> str:
        return f"CREATE CONSTRAINT ON (n:{self.item_label}) ASSERT EXISTS (n.{self.property})"

    def get_query_exist_drop(self) -> str:
        return f"DROP CONSTRAINT ON (n:{self.item_label}) ASSERT EXISTS (n.{self.property})"


TYPE_MAPPING = {
    "RelationshipStatus": GraphPropertyType.STRING,
    "BranchSupportType": GraphPropertyType.STRING,
    str: GraphPropertyType.STRING,
    int: GraphPropertyType.INTEGER,
    bool: GraphPropertyType.BOOLEAN,
    float: GraphPropertyType.FLOAT,
}


class ConstraintManagerBase:
    constraint_node_class: Optional[type[ConstraintItem]] = ConstraintItem
    constraint_rel_class: Optional[type[ConstraintItem]] = ConstraintItem

    def __init__(self, db: InfrahubDatabase):
        self.db = db

        self.nodes: list[ConstraintItem] = []
        self.rels: list[ConstraintItem] = []

    @property
    def items(self) -> list[ConstraintItem]:
        return self.nodes + self.rels

    @classmethod
    def from_graph_schema(cls, db: InfrahubDatabase, schema: dict[str, dict[str, BaseModel]]) -> Self:
        manager = cls(db=db)

        # Process the nodes first
        for schema_item in schema["nodes"].values():
            properties_class: BaseModel = schema_item.model_fields["properties"].annotation  # type: ignore[assignment]
            default_label = str(schema_item.model_fields["default_label"].default)
            for field_name, field in properties_class.model_fields.items():
                clean_field_name = field.alias or field_name
                item_type = None
                mandatory = True
                if isinstance(field.annotation, ForwardRef) and field.annotation.__forward_arg__ in TYPE_MAPPING:
                    item_type = TYPE_MAPPING[field.annotation.__forward_arg__]
                elif field.annotation in TYPE_MAPPING:
                    item_type = TYPE_MAPPING[field.annotation]
                elif get_origin(field.annotation) == Union and field.annotation.__args__[0] in TYPE_MAPPING:  # type: ignore[union-attr]
                    item_type = TYPE_MAPPING[field.annotation.__args__[0]]  # type: ignore[union-attr]
                    mandatory = False
                if not item_type or not manager.constraint_node_class:
                    continue

                manager.nodes.append(
                    manager.constraint_node_class(
                        item_name=default_label.lower(),
                        item_label=default_label,
                        property=clean_field_name,
                        type=item_type,
                        mandatory=mandatory,
                    )
                )

        # Process the relationships
        for label, schema_item in schema["relationships"].items():
            for field_name, field in schema_item.model_fields.items():
                clean_field_name = field.alias or field_name
                item_type = None
                mandatory = True
                if isinstance(field.annotation, ForwardRef) and field.annotation.__forward_arg__ in TYPE_MAPPING:
                    item_type = TYPE_MAPPING[field.annotation.__forward_arg__]
                elif field.annotation in TYPE_MAPPING:
                    item_type = TYPE_MAPPING[field.annotation]
                elif get_origin(field.annotation) == Union and field.annotation.__args__[0] in TYPE_MAPPING:  # type: ignore[union-attr]
                    item_type = TYPE_MAPPING[field.annotation.__args__[0]]  # type: ignore[union-attr]
                    mandatory = False
                if not item_type or not manager.constraint_rel_class:
                    continue

                manager.rels.append(
                    manager.constraint_rel_class(
                        item_name=label.lower(),
                        item_label=label,
                        property=clean_field_name,
                        type=item_type,
                        mandatory=mandatory,
                    )
                )

        return manager

    async def add(self) -> None:
        async with self.db.start_transaction() as dbt:
            for item in self.items:
                for query in item.get_add_queries():
                    await dbt.execute_query(query=query, params={}, name="constraint_add")

    async def drop(self) -> None:
        async with self.db.start_transaction() as dbt:
            for item in self.items:
                for query in item.get_drop_queries():
                    await dbt.execute_query(query=query, params={}, name="constraint_drop")

    async def list(self) -> list[ConstraintInfo]:
        raise NotImplementedError()


class ConstraintManagerNeo4j(ConstraintManagerBase):
    constraint_node_class = ConstraintNodeNeo4j
    constraint_rel_class = ConstraintRelNeo4j

    async def list(self) -> list[ConstraintInfo]:
        query = "SHOW CONSTRAINTS"
        records = await self.db.execute_query(query=query, params={}, name="constraint_show", type=QueryType.READ)
        results = []
        for record in records:
            results.append(
                ConstraintInfo(
                    item_name=record["name"],
                    item_label=", ".join(record["labelsOrTypes"]),
                    property=", ".join(record["properties"]),
                )
            )

        return results


class ConstraintManagerMemgraph(ConstraintManagerBase):
    constraint_node_class = ConstraintNodeMemgraph
    constraint_rel_class = None

    async def add(self) -> None:
        for item in self.items:
            for query in item.get_add_queries():
                await self.db.execute_query(query=query, params={}, name="constraint_add")

    async def drop(self) -> None:
        for item in self.items:
            for query in item.get_drop_queries():
                await self.db.execute_query(query=query, params={}, name="constraint_drop")

    async def list(self) -> list[ConstraintInfo]:
        query = "SHOW CONSTRAINT INFO"
        records = await self.db.execute_query(query=query, params={}, name="constraint_show", type=QueryType.READ)
        results = []
        for record in records:
            results.append(ConstraintInfo(item_name="n_a", item_label=record["label"], property=record["properties"]))

        return results
